# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import nemo_run as run
import pytest
import torch

from nemo.collections.llm.api import finetune, pretrain
from nemo.collections.llm.gpt.data.mock import MockDataModule
from nemo.collections.llm.gpt.data.squad import SquadDataModule
from nemo.collections.llm.gpt.model.starcoder import StarcoderConfig15B, StarcoderModel
from nemo.collections.llm.peft import PEFT_STR2CLS
from nemo.collections.llm.recipes import starcoder_15b
from nemo.lightning import Trainer


class TestStarcoder_15B:
    @pytest.fixture(scope="class")
    def recipe_module(self):
        return starcoder_15b

    def test_model(self, recipe_module):
        model_config = recipe_module.model()
        assert isinstance(model_config, run.Config)
        assert model_config.__fn_or_cls__ == StarcoderModel
        assert isinstance(model_config.config, run.Config)
        assert model_config.config.__fn_or_cls__ == StarcoderConfig15B

    def test_trainer(self, recipe_module):
        trainer_config = recipe_module.starcoder_trainer()
        assert isinstance(trainer_config, run.Config)
        assert trainer_config.__fn_or_cls__ == Trainer
        assert trainer_config.accelerator == "gpu"
        assert trainer_config.devices == 8
        assert trainer_config.num_nodes == 1
        assert trainer_config.max_steps == 1168251
        assert trainer_config.val_check_interval == 2000
        assert trainer_config.limit_test_batches == 32
        assert trainer_config.limit_val_batches == 32
        assert trainer_config.log_every_n_steps == 10

        # Check strategy configuration
        assert isinstance(trainer_config.strategy, run.Config)
        assert trainer_config.strategy.__fn_or_cls__.__name__ == "MegatronStrategy"
        assert trainer_config.strategy.tensor_model_parallel_size == 4
        assert trainer_config.strategy.pipeline_model_parallel_size == 2
        assert trainer_config.strategy.gradient_as_bucket_view is True
        assert trainer_config.strategy.ckpt_include_optimizer is True
        assert trainer_config.strategy.ckpt_async_save is True
        assert trainer_config.strategy.ckpt_parallel_load is True

    def test_pretrain_recipe(self, recipe_module):
        recipe = recipe_module.pretrain_recipe()
        assert isinstance(recipe, run.Partial)
        assert recipe.__fn_or_cls__ == pretrain
        assert isinstance(recipe.model, run.Config)
        assert recipe.model.__fn_or_cls__ == StarcoderModel
        assert isinstance(recipe.trainer, run.Config)
        assert recipe.trainer.__fn_or_cls__ == Trainer
        assert isinstance(recipe.data, run.Config)
        assert recipe.data.__fn_or_cls__ == MockDataModule
        assert recipe.data.seq_length == 4096
        assert recipe.data.global_batch_size == 32
        assert recipe.data.micro_batch_size == 2

        # Check parallelism settings
        assert recipe.trainer.strategy.tensor_model_parallel_size == 1
        assert recipe.trainer.strategy.pipeline_model_parallel_size == 8
        assert recipe.trainer.strategy.pipeline_dtype == torch.bfloat16

    def test_finetune_recipe(self, recipe_module):
        recipe = recipe_module.finetune_recipe()
        assert isinstance(recipe, run.Partial)
        assert recipe.__fn_or_cls__ == finetune
        assert isinstance(recipe.model, run.Config)
        assert recipe.model.__fn_or_cls__ == StarcoderModel
        assert isinstance(recipe.trainer, run.Config)
        assert recipe.trainer.__fn_or_cls__ == Trainer
        assert isinstance(recipe.data, run.Config)
        assert recipe.data.__fn_or_cls__ == SquadDataModule

        # Check PEFT configuration
        assert isinstance(recipe.peft, run.Config)
        assert recipe.peft.__fn_or_cls__ == PEFT_STR2CLS['lora']
        assert recipe.optim.config.lr == 1e-4

    def test_finetune_recipe_with_dora(self, recipe_module):
        recipe = recipe_module.finetune_recipe(peft_scheme='dora')
        assert isinstance(recipe.peft, run.Config)
        assert recipe.peft.__fn_or_cls__ == PEFT_STR2CLS['dora']
        assert recipe.optim.config.lr == 1e-4

    def test_finetune_recipe_without_peft(self, recipe_module):
        recipe = recipe_module.finetune_recipe(peft_scheme=None)
        assert not hasattr(recipe, 'peft') or recipe.peft is None
        assert recipe.trainer.strategy.pipeline_model_parallel_size == 8
        assert recipe.optim.config.lr == 5e-6

    def test_finetune_recipe_with_invalid_peft(self, recipe_module):
        with pytest.raises(ValueError, match="Unrecognized peft scheme: invalid_scheme"):
            recipe_module.finetune_recipe(peft_scheme="invalid_scheme")



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import shutil
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

from datasets import DatasetDict, load_dataset

from nemo.collections.llm.gpt.data.core import get_dataset_root
from nemo.collections.llm.gpt.data.fine_tuning import FineTuningDataModule
from nemo.lightning.io.mixin import IOMixin
from nemo.utils import logging

if TYPE_CHECKING:
    from nemo.collections.common.tokenizers import TokenizerSpec
    from nemo.collections.llm.gpt.data.packed_sequence import PackedSequenceSpecs


class SquadDataModule(FineTuningDataModule, IOMixin):
    """A data module for fine-tuning on the Squad dataset.

    This class inherits from the `FineTuningDataModule` class and is specifically designed for
    fine-tuning models on the Stanford Question Answering Dataset (SQuAD). It handles data download,
    preprocessing, splitting, and preparing the data in a format suitable for training,
    validation, and testing.

    Args:
        dataset_root (Optional[Union[str, Path]]): The root directory containing the training,
            validation, and test data. Defaults to None, which by default downloads the data.
        force_redownload (bool, optional): Whether to force re-download the dataset even if it
            exists locally. Defaults to False.
        delete_raw (bool, optional): Whether to delete the raw downloaded dataset after preprocessing.
            Defaults to True.

        See FineTuningDataModule for the other args
    """

    def __init__(
        self,
        dataset_root: Optional[Union[str, Path]] = None,
        seq_length: int = 2048,
        tokenizer: Optional["TokenizerSpec"] = None,
        micro_batch_size: int = 4,
        global_batch_size: int = 8,
        rampup_batch_size: Optional[List[int]] = None,
        force_redownload: bool = False,
        delete_raw: bool = True,
        seed: int = 1234,
        memmap_workers: int = 1,
        num_workers: int = 8,
        pin_memory: bool = True,
        persistent_workers: bool = False,
        packed_sequence_specs: Optional["PackedSequenceSpecs"] = None,
        dataset_kwargs: Optional[Dict[str, Any]] = None,
    ):
        self.force_redownload = force_redownload
        self.delete_raw = delete_raw

        super().__init__(
            dataset_root=dataset_root if dataset_root is not None else get_dataset_root("squad"),
            seq_length=seq_length,
            tokenizer=tokenizer,
            micro_batch_size=micro_batch_size,
            global_batch_size=global_batch_size,
            rampup_batch_size=rampup_batch_size,
            seed=seed,
            memmap_workers=memmap_workers,
            num_workers=num_workers,
            pin_memory=pin_memory,
            persistent_workers=persistent_workers,
            packed_sequence_specs=packed_sequence_specs,
            dataset_kwargs=dataset_kwargs,
        )

    def prepare_data(self) -> None:
        # if train file is specified, no need to do anything
        if not self.train_path.exists() or self.force_redownload:
            dset = self._download_data()
            self._preprocess_and_split_data(dset)
        super().prepare_data()

    def _download_data(self):
        logging.info(f"Downloading {self.__class__.__name__}...")
        return load_dataset(
            "squad",
            cache_dir=str(self.dataset_root),
            download_mode="force_redownload" if self.force_redownload else None,
        )

    def _preprocess_and_split_data(
        self, dset: DatasetDict, split_val_from_train: bool = True, val_proportion: float = 0.05
    ):
        """Preprocesses and splits the downloaded dataset into training, validation, and test sets.

        Args:
            dset (DatasetDict): The downloaded dataset object.
            split_val_from_train (bool, optional): Whether to split the validation set from the training set.
                If False, the validation set is split from the test set. Defaults to True.
            val_proportion (float, optional): The proportion of the training or test set to be used
                for the validation split. Defaults to 0.05.
        """
        logging.info(f"Preprocessing {self.__class__.__name__} to jsonl format and splitting...")
        save_splits = {}
        train_set = dset.get('train')
        val_set = dset.get('validation')

        if split_val_from_train:
            split_dataset = train_set.train_test_split(test_size=val_proportion, seed=self.seed)
            save_splits['training'] = split_dataset['train']
            save_splits['validation'] = split_dataset['test']
            save_splits['test'] = val_set
        else:
            split_dataset = val_set.train_test_split(test_size=val_proportion, seed=self.seed)
            save_splits['training'] = train_set
            save_splits['validation'] = split_dataset['test']
            save_splits['test'] = split_dataset['train']

        for split_name, dataset in save_splits.items():
            output_file = self.dataset_root / f"{split_name}.jsonl"

            with output_file.open("w", encoding="utf-8") as f:
                for example in dataset:
                    json_line = {}
                    # Write each example as a JSON line in the output file
                    json_line["input"] = (
                        "Context: " + example["context"] + " Question: " + example['question'] + " Answer:"
                    )
                    json_line["output"] = example["answers"]["text"][0]
                    if split_name == "test":
                        json_line["original_answers"] = example["answers"]["text"]
                    f.write(json.dumps(json_line) + "\n")

            logging.info(f"{split_name} split saved to {output_file}")

        if self.delete_raw:
            for p in self.dataset_root.iterdir():
                if p.is_dir():
                    shutil.rmtree(p)
                elif '.jsonl' not in str(p.name):
                    p.unlink()

    def reconfigure_limit_batches(self):
        """no op"""
        return



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from typing import List, Optional, Tuple, Union

from nemo.collections.common.parts.perf_metrics_utils import LLM_VOCAB_SIZE_MAP


@dataclass
class FLOPSConfig:
    """Contains the model hparams needed for FLOPS computations"""

    gbs: int
    enc_seq_len: Optional[int] = None
    hs: Optional[int] = None
    layers: Optional[int] = None
    ffn_hs: Optional[int] = None
    attention_heads: Optional[int] = None
    moe_router_topk: Optional[int] = None
    query_groups: Optional[int] = None
    kv_channels: Optional[int] = None
    img_seq_len: Optional[int] = None
    img_h: Optional[int] = None
    img_w: Optional[int] = None
    in_channels: Optional[int] = None
    patch_dim: Optional[int] = None
    class_token_len: Optional[int] = None
    projector_type: Optional[str] = None
    inp_s: Optional[int] = None
    model_pattern: Optional[str] = None
    vocab_size: Optional[int] = None
    model_channels: Optional[int] = None
    vec_in_dim: Optional[int] = None
    q_lora_rank: Optional[int] = None
    kv_lora_rank: Optional[int] = None
    qk_head_dim: Optional[int] = None
    qk_pos_emb_head_dim: Optional[int] = None
    v_head_dim: Optional[int] = None
    moe_layer_freq: Union[int, List[int]] = None
    moe_shared_expert_intermediate_size: Optional[int] = None
    moe_ffn_hidden_size: Optional[int] = None
    mtp_num_layers: Optional[int] = None
    causal_self_attn: Optional[bool] = None
    is_hybrid_model: bool = False
    hybrid_override_pattern: Optional[str] = None
    mamba_state_dim: Optional[int] = None
    mamba_head_dim: Optional[int] = None
    mamba_num_groups: Optional[int] = None
    mamba_num_heads: Optional[int] = None
    # SWA configs
    window_attn_skip_freq: Optional[Union[int, List[int]]] = None
    window_size: Optional[Tuple[int, int]] = (128, 0)


def gpt3(config: FLOPSConfig):
    """Model FLOPs for GPT3 family"""
    vocab_size = LLM_VOCAB_SIZE_MAP["gpt3"]
    causal_self_attn = True

    return (
        24 * config.gbs * config.enc_seq_len * config.hs * config.hs
        + 4 * config.gbs * config.enc_seq_len * config.enc_seq_len * config.hs * (0.5 if causal_self_attn else 1)
    ) * (3 * config.layers) + (6 * config.gbs * config.enc_seq_len * config.hs * vocab_size)


def llama2(config: FLOPSConfig):
    """Model FLOPs for llama2 family"""
    vocab_size = LLM_VOCAB_SIZE_MAP["llama2"]
    causal_self_attn = True

    return (
        config.gbs
        * config.enc_seq_len
        * config.layers
        * config.hs
        * config.hs
        * (
            12
            + (12 * config.query_groups / config.attention_heads)
            + (18 * config.ffn_hs / config.hs)
            + (12 * config.enc_seq_len / config.hs) * (0.5 if causal_self_attn else 1)
            + (6 * vocab_size / (config.layers * config.hs))
        )
    )


def llama3(config: FLOPSConfig):
    """Model FLOPs for llama3 family"""
    vocab_size = LLM_VOCAB_SIZE_MAP["llama3"]
    causal_self_attn = True

    return (
        config.gbs
        * config.enc_seq_len
        * config.layers
        * config.hs
        * config.hs
        * (
            12
            + (12 * config.query_groups / config.attention_heads)
            + (18 * config.ffn_hs / config.hs)
            + (12 * config.enc_seq_len / config.hs) * (0.5 if causal_self_attn else 1)
            + (6 * vocab_size / (config.layers * config.hs))
        )
    )


def nemotron(config: FLOPSConfig):
    """Model FLOPs for nemotron family"""
    vocab_size = LLM_VOCAB_SIZE_MAP["nemotron"]
    causal_self_attn = True

    return (
        config.gbs
        * config.enc_seq_len
        * config.layers
        * config.hs
        * config.hs
        * (
            12
            + (12 * config.query_groups / config.attention_heads)
            + (12 * config.ffn_hs / config.hs)
            + (12 * config.enc_seq_len / config.hs) * (0.5 if causal_self_attn else 1)
            + (6 * vocab_size / (config.layers * config.hs))
        )
    )


def mixtral(config: FLOPSConfig):
    """Model FLOPs for mixtral family"""
    vocab_size = LLM_VOCAB_SIZE_MAP["mixtral"]
    causal_self_attn = True

    return (
        config.gbs
        * config.enc_seq_len
        * config.layers
        * config.hs
        * config.hs
        * (
            12
            + (12 * config.query_groups / config.attention_heads)
            + (18 * config.moe_router_topk * config.ffn_hs / config.hs)
            + (12 * config.enc_seq_len / config.hs) * (0.5 if causal_self_attn else 1)
            + (6 * vocab_size / (config.layers * config.hs))
        )
    )


def qwen3(config: FLOPSConfig):
    """Model FLOPs for Qwen3 family"""
    causal_self_attn = True
    seq_len = config.enc_seq_len
    hidden_size = config.hs
    gated_linear_multiplier = 2
    query_projection_to_hidden_size_ratio = config.kv_channels * config.attention_heads / hidden_size

    # attention flops for GQA
    attention_flops = (
        3
        * 2
        * config.gbs
        * config.layers
        * seq_len
        * hidden_size
        * hidden_size
        * query_projection_to_hidden_size_ratio
        * (
            (config.query_groups / config.attention_heads * 2 + 1)  # QKV gemm
            + (seq_len / hidden_size * 2 * (0.5 if causal_self_attn else 1))  # attention
            + 1  # attention proj gemm
        )
    )

    mlp_ffn_hidden_size = config.ffn_hs
    if hasattr(config, "moe_ffn_hidden_size") and config.moe_ffn_hidden_size is not None:
        mlp_ffn_hidden_size = config.moe_ffn_hidden_size * config.moe_router_topk

    # mlp flops
    mlp_flops = (
        3
        * 2
        * config.gbs
        * config.layers
        * seq_len
        * hidden_size
        * (1 + gated_linear_multiplier)
        * mlp_ffn_hidden_size
    )

    # vocab flops
    vocab_flops = 3 * 2 * config.gbs * seq_len * hidden_size * config.vocab_size

    return attention_flops + mlp_flops + vocab_flops


def bert(config: FLOPSConfig):
    """Model FLOPs for BERT family"""
    vocab_size = LLM_VOCAB_SIZE_MAP["bert"]

    return (
        72
        * config.gbs
        * config.layers
        * config.enc_seq_len
        * config.hs
        * config.hs
        * (1 + (config.enc_seq_len / (6 * config.hs)) + (vocab_size / (12 * config.hs * config.layers)))
    )


def transformer(config: FLOPSConfig):
    """Calculate FLOPs for a standard Transformer model.
    Note: This does not cover encoder-decoder models.
    """
    # Extract parameters from config
    batch_size = config.gbs
    hidden_size = config.hs
    seq_length = config.enc_seq_len
    num_layers = config.layers
    num_attention_heads = config.attention_heads
    ffn_hidden_size = config.ffn_hs
    vocab_size = config.vocab_size

    if vocab_size is None:
        raise ValueError("vocab_size is required for transformer FLOPs calculation")

    # Handle optional parameters with reasonable defaults
    query_groups = config.query_groups if config.query_groups is not None else num_attention_heads
    causal_self_attn = config.causal_self_attn if config.causal_self_attn is not None else False
    moe_router_topk = config.moe_router_topk if config.moe_router_topk is not None else 0
    kv_channels = hidden_size // num_attention_heads  # Standard dimension per head

    # Calculate query projection size and ratio
    query_projection_size = kv_channels * num_attention_heads
    query_projection_to_hidden_size_ratio = query_projection_size / hidden_size

    # MoE parameters - simplified for NeMo config
    # In this implementation, we assume all layers are dense if num_experts is None
    if moe_router_topk == 0:
        num_dense_layers = num_layers
        num_moe_layers = 0
        num_experts_routed_to = 0
    else:
        # Simplified MoE handling - assuming uniform distribution of MoE layers
        # This can be expanded based on NeMo's actual MoE implementation
        num_moe_layers = num_layers // 2  # Simplified assumption
        num_dense_layers = num_layers - num_moe_layers
        num_experts_routed_to = moe_router_topk

    # Handle SwiGLU vs standard GELU/ReLU
    # Default to standard activation (no SwiGLU)
    gated_linear_multiplier = 1

    # Define the expansion factor as described in the paper
    # 3x: Each GEMM needs forward pass, backward wgrad, and backward dgrad
    # 2x: GEMMs are stacked twice in standard Transformer architectures
    # 2x: A GEMM of m*n with n*k requires 2mnk floating-point operations
    expansion_factor = 3 * 2 * 2
    # Attention
    if not causal_self_attn:
        attention_component = (
            1
            + (query_groups / num_attention_heads)
            # Only half of the attention matrix is non-zero and needs to be multiplied with V
            + (seq_length / hidden_size)  # If causal self attn -> divide by 2.
        ) * query_projection_to_hidden_size_ratio
    else:
        attention_component = (
            1
            + (query_groups / num_attention_heads)
            # Only half of the attention matrix is non-zero and needs to be multiplied with V
            + (seq_length / hidden_size / 2)  # If causal self attn -> divide by 2.
        ) * query_projection_to_hidden_size_ratio

    # Calculate total FLOPs
    total_flops = (
        expansion_factor
        * batch_size
        * seq_length
        * num_layers
        * hidden_size
        * hidden_size
        * (
            attention_component
            # MLP component
            + (
                (
                    # Dense layers
                    (ffn_hidden_size * num_dense_layers)
                    +
                    # MoE layers
                    (
                        (
                            # Routed experts
                            ffn_hidden_size
                            * num_experts_routed_to
                            # Note: Shared experts are not implemented in this version
                        )
                        * num_moe_layers
                    )
                )
                * gated_linear_multiplier
                / (num_layers * hidden_size)
            )
            # Logit component
            + (vocab_size / (2 * num_layers * hidden_size))
        )
    )

    return total_flops


def clip_vit_l(config: FLOPSConfig):
    """Model FLOPs for CLIP ViT"""

    if config.img_seq_len is None:
        config.img_seq_len = (config.img_h * config.img_w) / (
            config.patch_dim * config.patch_dim
        ) + config.class_token_len
    return config.gbs * config.layers * config.hs * config.hs * config.img_seq_len * (
        24 + (4 * config.img_seq_len / config.hs)
    ) + (2 * config.gbs * config.hs * config.in_channels * config.img_h * config.img_w)


def neva_projection(config: FLOPSConfig):
    """Model FLOPs for NeVA Projection"""

    if "mlp" in config.projector_type:
        return 6 * config.gbs * config.img_seq_len * config.ffn_hs * (config.inp_s + config.hs)
    elif config.projector_type == "affine":
        return 6 * config.gbs * config.img_seq_len * config.inp_s * config.hs
    else:
        raise ValueError(
            f"NeVA Projections FLOPs calculator only supports 'mlp', 'mcore_mlp'"
            f" or 'affine' projector_type but found {config.projector_type}"
        )


def flux(config: FLOPSConfig):
    """Model FLOPs for FLUX"""

    hs = config.hs
    seq_len = config.model_channels + config.inp_s
    base_factor = 6 * config.gbs  # common multiplier for most terms

    # Joint layer computations
    joint_layer_flops = (
        base_factor
        * config.layers[0]
        * (
            10 * hs * hs  # hidden size operations
            + 2 * hs * (config.model_channels + config.inp_s) * (1 + hs * 7)  # channel and context joint attention
            + 2 * (config.model_channels + config.inp_s) * hs  # final projection
        )
    )

    # Single layer computations
    single_layer_flops = (
        base_factor
        * config.layers[1]
        * seq_len
        * hs
        * (
            3  # linear Y
            + 1  # Modulation
            + 4 * hs  # Linear computations
            + (3 * hs + 2 * seq_len)  # attention operations
            + 5 * hs  # feed-forward
            + 1  # Modulation
        )
    )

    # Embedding and projection layers
    other_flops = base_factor * (
        config.inp_s * config.in_channels * hs  # image embedding
        + config.inp_s * hs * config.model_channels  # text embedding
        + config.vec_in_dim * hs
        + hs * hs  # vector embedding
        + 2 * (config.model_channels * hs + hs * hs)  # guidance + timestep embedding
        + (config.inp_s * config.in_channels * hs) / config.gbs  # final projection
    )

    return joint_layer_flops + single_layer_flops + other_flops


def deepseekv3(config: FLOPSConfig):
    """Model FLOPs for DeepSeek V3"""

    # self-attention flops
    bmm1_flops = (
        0.5 * (config.qk_head_dim + config.qk_pos_emb_head_dim) * config.attention_heads * (config.enc_seq_len**2)
    )
    bmm2_flops = 0.5 * config.v_head_dim * config.attention_heads * (config.enc_seq_len**2)
    per_input_attention_flops = 6 * (bmm1_flops + bmm2_flops) * config.layers
    if config.mtp_num_layers is not None:
        per_input_attention_flops += 6 * (bmm1_flops + bmm2_flops) * config.mtp_num_layers

    # linear layer flops
    per_layer_mla_params = config.hs * config.q_lora_rank + config.q_lora_rank * (
        (config.qk_head_dim + config.qk_pos_emb_head_dim) * config.attention_heads
    )  # Q
    per_layer_mla_params += config.hs * config.qk_pos_emb_head_dim  # K^R
    per_layer_mla_params += config.hs * config.kv_lora_rank + config.kv_lora_rank * (
        (config.qk_head_dim + config.v_head_dim) * config.attention_heads
    )  # K^C and V^C
    per_layer_mla_params += config.v_head_dim * config.attention_heads * config.hs  # Proj
    mla_params = per_layer_mla_params * config.layers
    if config.mtp_num_layers is not None:
        mla_params += per_layer_mla_params * config.mtp_num_layers

    dense_layer_ffn_params = config.hs * config.ffn_hs * 3  # gated linear unit
    per_shared_expert_params = config.hs * config.moe_shared_expert_intermediate_size * 3
    per_selected_expert_params = config.hs * config.moe_ffn_hidden_size * 3
    ffn_params = 0

    if isinstance(config.moe_layer_freq, int):
        moe_layer_pattern = [1 if (i % config.moe_layer_freq == 0) else 0 for i in range(config.layers)]
    else:
        moe_layer_pattern = config.moe_layer_freq
    for i in moe_layer_pattern:
        if i == 0:
            ffn_params += dense_layer_ffn_params
        else:
            ffn_params += per_shared_expert_params + (per_selected_expert_params * config.moe_router_topk)
    if config.mtp_num_layers is not None:
        for i in range(config.mtp_num_layers):
            ffn_params += per_shared_expert_params + (per_selected_expert_params * config.moe_router_topk)
    per_input_params = mla_params + ffn_params
    per_input_linear_flops = 6 * per_input_params * config.enc_seq_len

    # vocab flops
    per_input_vocab_flops = 6 * config.vocab_size * config.hs * config.enc_seq_len
    if config.mtp_num_layers is not None:
        for i in range(config.mtp_num_layers):
            per_input_vocab_flops += 6 * config.vocab_size * config.hs * config.enc_seq_len
            per_input_vocab_flops += 6 * config.hs * 2 * config.hs * config.enc_seq_len

    return (per_input_attention_flops + per_input_linear_flops + per_input_vocab_flops) * config.gbs


def _nemotronh_mlp_layer_flops(config: FLOPSConfig):
    """Model FLOPs for MLP layer. Assume gated linear unit."""
    return 6 * config.gbs * config.enc_seq_len * config.hs * config.ffn_hs * 3


def _non_mla_attn_layer_flops(config: FLOPSConfig):
    """Model FLOPs for attention layer"""
    return (
        6
        * config.gbs
        * config.enc_seq_len
        * config.hs
        * (
            config.hs  # Q
            + config.query_groups / config.attention_heads * config.hs * 2  # KV
            + config.enc_seq_len / 2 * 2
            + config.hs
        )
    )


def _mamba_layer_flops(config: FLOPSConfig):
    """Model FLOPs for Mamba layer. We ignore part of the flops of scan because the
    chunk size is not known from model config."""
    assert config.mamba_state_dim is not None
    assert config.mamba_head_dim is not None

    if config.mamba_num_heads:
        nheads = config.mamba_num_heads
    else:
        nheads = 2 * config.hs // config.mamba_head_dim  # default expand is 2
    d_in = nheads * config.mamba_head_dim
    return (
        (
            6
            * config.gbs
            * config.enc_seq_len
            * config.hs
            * (2 * d_in + 2 * config.mamba_num_groups * config.mamba_state_dim + nheads)
        )
        + (3 * 2 * config.gbs * config.enc_seq_len * d_in * config.mamba_state_dim)
        + (6 * config.gbs * config.enc_seq_len * d_in * config.hs)
    )


def _hybrid_model_flops(config: FLOPSConfig):
    """Model FLOPs for hybrid model"""
    assert config.is_hybrid_model == True
    assert config.hybrid_override_pattern is not None

    num_attn_layers, num_mamba_layers, num_mlp_layers = 0, 0, 0
    for c in config.hybrid_override_pattern:
        if c == 'M':
            num_mamba_layers += 1
        elif c == '-':
            num_mlp_layers += 1
        elif c == '*':
            num_attn_layers += 1
    return (
        num_attn_layers * _non_mla_attn_layer_flops(config)
        + num_mamba_layers * _mamba_layer_flops(config)
        + num_mlp_layers * _nemotronh_mlp_layer_flops(config)
        + 6 * config.gbs * config.enc_seq_len * config.hs * config.vocab_size
    )


def nemotronh(config: FLOPSConfig):
    """Model FLOPs for NemotronH"""
    return _hybrid_model_flops(config)


def attention_flops_calculator(
    seqlen,
    hidden_size,
    num_attention_heads,
    num_query_groups,
    kv_channels: Optional[int] = None,
    is_swa: bool = False,
    swa_window_size: int = 128,
):
    """Calculate the flops for the attention part."""
    kv_channels = kv_channels or (hidden_size // num_attention_heads)

    linear_qkv = seqlen * hidden_size * (kv_channels * (num_attention_heads + num_query_groups * 2))

    linear_proj = seqlen * hidden_size * (kv_channels * num_attention_heads)

    if is_swa:
        attention_mask_nz_elem = (
            swa_window_size * (swa_window_size + 1) / 2 + (seqlen - swa_window_size) * swa_window_size
        )
        attention = num_attention_heads * (attention_mask_nz_elem * kv_channels) * 2
    else:
        bmm_k = kv_channels
        bmm_b = num_attention_heads
        attention_mask_nz_elem = seqlen * (seqlen + 1) / 2
        attention = bmm_b * attention_mask_nz_elem * bmm_k * 2

    return (linear_qkv + linear_proj + attention) * 6


def moe_mlp_flops_calculator(
    seqlen,
    hidden_size,
    moe_ffn_hidden_size,
    moe_router_topk,
    gated_linear_unit: bool = True,
):
    """Calculate the flops for the MLP"""
    total_num_tokens = seqlen * moe_router_topk
    linear_fc1 = total_num_tokens * hidden_size * moe_ffn_hidden_size * (2 if gated_linear_unit else 1)
    linear_fc2 = total_num_tokens * moe_ffn_hidden_size * hidden_size
    return (linear_fc1 + linear_fc2) * 6


def loss_flops_calculator(
    seqlen,
    hidden_size,
    vocab_size,
):
    """Calculate the flops for the loss"""
    return (seqlen * hidden_size * vocab_size) * 6


def gpt_oss_flops_calculator(
    gbs,
    num_layers,
    seqlen,
    hidden_size,
    num_attention_heads,
    num_query_groups,
    moe_ffn_hidden_size,
    moe_router_topk,
    vocab_size,
    kv_channels: Optional[int] = None,
    swa_window_size: int = 128,
    window_attn_skip_freq: Optional[int] = 2,
):
    """Calculate the flops for the GPT-OSS model"""
    flops = 0
    for i in range(num_layers):
        if i % window_attn_skip_freq == 0:
            flops += attention_flops_calculator(
                seqlen,
                hidden_size,
                num_attention_heads,
                num_query_groups,
                kv_channels,
                is_swa=False,
            )
        else:
            flops += attention_flops_calculator(
                seqlen,
                hidden_size,
                num_attention_heads,
                num_query_groups,
                kv_channels,
                is_swa=True,
                swa_window_size=swa_window_size,
            )
        flops += moe_mlp_flops_calculator(
            seqlen,
            hidden_size,
            moe_ffn_hidden_size,
            moe_router_topk,
        )
    flops += loss_flops_calculator(seqlen, hidden_size, vocab_size)
    flops *= gbs
    return flops


def gpt_oss(config: FLOPSConfig):
    """Model FLOPs for GPT-OSS"""
    return gpt_oss_flops_calculator(
        gbs=config.gbs,
        num_layers=config.layers,
        seqlen=config.enc_seq_len,
        hidden_size=config.hs,
        num_attention_heads=config.attention_heads,
        num_query_groups=config.query_groups,
        moe_ffn_hidden_size=config.moe_ffn_hidden_size,
        moe_router_topk=config.moe_router_topk,
        vocab_size=config.vocab_size,
        kv_channels=config.kv_channels,
        swa_window_size=config.window_size[0] if config.window_size is not None else 128,
        window_attn_skip_freq=config.window_attn_skip_freq,
    )



# Copyright (c) 2022, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC
from typing import List, Optional, Tuple

import torch

from nemo.collections.asr.parts.k2.grad_utils import GradExpNormalize
from nemo.collections.asr.parts.k2.utils import (
    create_supervision,
    get_arc_weights,
    get_uniform_rnnt_prune_ranges,
    make_non_pad_mask,
    make_non_pad_mask_3d,
    prep_padded_densefsavec,
)
from nemo.core.utils.k2_guard import k2  # import k2 from guard module


class CtcK2Mixin(ABC):
    """k2 Mixin class that simplifies the construction of various k2-based CTC-like losses.
    
    It does the following:
        -   Prepares and adapts the input tensors (method _prepare_log_probs_and_targets).
        -   Creates Emissions graphs (method _prepare_emissions_graphs).
        -   Extracts the labels and probabilities of the best lattice path (method _extract_labels_and_probabilities).
    """

    def _prepare_log_probs_and_targets(
        self,
        log_probs: torch.Tensor,
        input_lengths: torch.Tensor,
        targets: Optional[torch.Tensor] = None,
        target_lengths: Optional[torch.Tensor] = None,
    ) -> Tuple[torch.Tensor, torch.Tensor, Optional[torch.Tensor], Optional[torch.Tensor]]:
        """Creates k2-style supervisions and shifts targets by one if the <blank> number is not zero.
        """
        assert log_probs.size(-1) == self.num_classes
        supervisions = create_supervision(input_lengths)
        # shift targets to make output epsilon ID zero
        return (
            log_probs,
            supervisions,
            torch.where(targets < self.blank, targets + 1, targets) if targets is not None else None,
            target_lengths,
        )

    def _prepare_emissions_graphs(self, log_probs: torch.Tensor, supervisions: torch.Tensor) -> 'k2.DenseFsaVec':
        """Creates DenseFsaVec, padding it with <epsilon> frames if the topology is `compact`.
        In particular, every second frame of the DenseFsaVec is the <epsilon> frame.

        <epsilon> frame is a frame with <epsilon> log-probability zero and every other log-probability is -inf.
        """
        return (
            prep_padded_densefsavec(log_probs, supervisions)
            if self.pad_fsavec
            else k2.DenseFsaVec(log_probs, supervisions)
        )

    def _maybe_normalize_gradients(self, log_probs: torch.Tensor, input_lengths: torch.Tensor) -> torch.Tensor:
        """PyTorch is doing the log-softmax normalization as part of the CTC computation.
        More: https://github.com/k2-fsa/k2/issues/575
        """
        return GradExpNormalize.apply(log_probs, input_lengths, "mean" if self.reduction != "sum" else "none")

    def _extract_labels_and_probabilities(
        self, shortest_path_fsas: 'k2.Fsa', return_ilabels: bool = False, output_aligned: bool = True
    ) -> Tuple[List[torch.Tensor], List[torch.Tensor]]:
        """Extracts the labels and probabilities of the best lattice path,
        dropping <epsilon> arcs and restoring the targets shift, if needed.
        """
        shortest_paths = []
        probs = []
        # direct iterating does not work as expected
        for i in range(shortest_path_fsas.shape[0]):
            shortest_path_fsa = shortest_path_fsas[i]
            # suppose that artificial input epsilon numbers >= self.num_classes
            non_eps_mask = (shortest_path_fsa.labels != -1) & (shortest_path_fsa.labels < self.num_classes)
            if return_ilabels:
                labels = shortest_path_fsa.labels[non_eps_mask]
            else:
                labels = shortest_path_fsa.aux_labels[non_eps_mask]
                if self.blank != 0:
                    # suppose output epsilon number == 0
                    # since the input epsilons were removed, we treat all remaining epsilons as blanks
                    labels[labels == 0] = self.blank
                    labels[(labels > 0) & (labels < self.blank)] -= 1
            labels = labels.to(dtype=torch.long)
            if not return_ilabels and not output_aligned:
                labels = labels[labels != self.blank]
            shortest_paths.append(labels)
            probs.append(get_arc_weights(shortest_path_fsa)[non_eps_mask].exp().to(device=shortest_path_fsas.device))
        return shortest_paths, probs


class RnntK2Mixin(CtcK2Mixin):
    """k2 Mixin class that simplifies the construction of various k2-based RNNT-like losses. Inherits CtcK2Mixin.
    
    It does the following:
        -   Prepares and adapts the input tensors.
        -   Creates Emissions graphs.
        -   Extracts the labels and probabilities of the best lattice path (method _extract_labels_and_probabilities).
    """

    def _prepare_log_probs_and_targets(
        self,
        log_probs: torch.Tensor,
        input_lengths: torch.Tensor,
        targets: torch.Tensor,
        target_lengths: torch.Tensor,
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
        """Before calling super()._prepare_log_probs_and_targets, this method reshapes the log_probs tensor 
        from (B, T, U+1, D) to (B, T', D) where T' = T*(U+1), shifts paddings along T and U towards the end of T', 
        and recomputes input_lengths.

        It also calculates indices on which <epsilon> steps should be applied to the log_probs tensor to emulate 
        <blank> arcs shift of the Emissions graph for the pruned RNNT variant.
        """
        assert len(log_probs.size()) == 4  # B T U D
        B, T, U, D = log_probs.size()
        TU = T * U

        # save step indices if, as we assume, decoder output pruning has been applied
        if self.predictor_window_size > 0 and self.predictor_window_size < target_lengths.max():
            window_size_with_blank = self.predictor_window_size + 1
            ranges_begin = get_uniform_rnnt_prune_ranges(
                input_lengths, target_lengths, window_size_with_blank, self.predictor_step_size, T, True
            )
            step_sizes = ranges_begin[:, 1:] - ranges_begin[:, :-1]
            raw_step_indices = torch.where(step_sizes > 0)
            if self.predictor_step_size > 1:
                raw_step_indices = torch.repeat_interleave(
                    torch.stack(raw_step_indices).T, step_sizes[raw_step_indices], dim=0
                ).T
                raw_step_indices = (raw_step_indices[0], raw_step_indices[1])
            unique, count = torch.unique(raw_step_indices[0], return_counts=True)
            shift_mask = raw_step_indices[0].unsqueeze(0).repeat(len(unique), 1) == unique.unsqueeze(-1)
            step_indices = (
                raw_step_indices[0],
                (
                    torch.arange(ranges_begin.size(1)).unsqueeze(0).repeat(ranges_begin.size(0), 1)
                    * window_size_with_blank
                )[(raw_step_indices[0], raw_step_indices[1] + 1)]
                + torch.cumsum(shift_mask, 1)[shift_mask]
                - 1,
            )
            max_count = count.max()
            max_count_vec = torch.full((B,), max_count)
            max_count_vec[unique] -= count
            pad_indices_row = torch.repeat_interleave(torch.arange(B), max_count_vec)
            pad_unique = torch.unique(pad_indices_row)
            pad_shift_mask = pad_indices_row.unsqueeze(0).repeat(len(pad_unique), 1) == pad_unique.unsqueeze(-1)
            pad_indices = (
                pad_indices_row,
                T * window_size_with_blank + max_count - torch.cumsum(pad_shift_mask, 1)[pad_shift_mask],
            )
            self.__step_indices = (
                torch.cat((step_indices[0], pad_indices[0])),
                torch.cat((step_indices[1], pad_indices[1])),
            )
            self.__supervisions_add = max_count - max_count_vec
        else:
            self.__step_indices = None
            self.__supervisions_add = None

        # reshape 4D log_probs to 3D with respect to target_lengths
        non_pad_mask_true = make_non_pad_mask_3d(input_lengths, target_lengths + 1, T, U).flatten(1)
        input_lengths = non_pad_mask_true.sum(1)
        non_pad_mask_fake = make_non_pad_mask(input_lengths, TU).flatten()
        non_pad_mask_true = non_pad_mask_true.flatten()
        rearranged_indices = torch.arange(TU * B, device=log_probs.device)
        rearranged_indices_buffer = rearranged_indices.clone()
        rearranged_indices[non_pad_mask_fake] = rearranged_indices_buffer[non_pad_mask_true]
        rearranged_indices[~non_pad_mask_fake] = rearranged_indices_buffer[~non_pad_mask_true]
        log_probs = log_probs.reshape(-1, D)[rearranged_indices].view(B, -1, D)

        return super()._prepare_log_probs_and_targets(log_probs, input_lengths, targets, target_lengths)

    def _prepare_emissions_graphs(self, log_probs: torch.Tensor, supervisions: torch.Tensor) -> 'k2.DenseFsaVec':
        """Overrides super()._prepare_emissions_graphs.
        Creates DenseFsaVec, adding <epsilon> outputs to the end of the D dimension.

        If pruning is used, this method also pads the DenseFsaVec with <epsilon> frames 
        according to the <epsilon> steps, calculated before.

        <epsilon> frame is a frame with <epsilon> log-probability zero and every other log-probability is -inf.
        """
        if self.__step_indices is None or self.__supervisions_add is None:
            log_probs_eps = torch.cat(
                (log_probs, torch.zeros((log_probs.size(0), log_probs.size(1), 1), device=log_probs.device)), dim=2
            )
        else:
            mask = torch.zeros(
                (log_probs.size(0), log_probs.size(1) + int(len(self.__step_indices[0]) / log_probs.size(0))),
                dtype=torch.bool,
            )
            mask[self.__step_indices] = True
            log_probs_eps = torch.zeros((mask.size(0), mask.size(1), log_probs.size(2) + 1), device=log_probs.device)
            log_probs_eps[mask] = torch.tensor(
                [torch.finfo(torch.float32).min] * log_probs.size(2) + [0], device=log_probs.device
            )
            log_probs_eps[~mask] = torch.cat(
                (log_probs, torch.zeros((log_probs.size(0), log_probs.size(1), 1), device=log_probs.device)), dim=2
            ).view(-1, log_probs.size(-1) + 1)
            input_lengths = supervisions[:, -1] + self.__supervisions_add[supervisions[:, 0].to(dtype=torch.long)]
            if not torch.all(input_lengths[:-1] - input_lengths[1:] >= 0):
                # have to reorder supervisions inplace
                order = torch.argsort(input_lengths, descending=True)
                # the second column is assumed to be zero
                supervisions[:, 0] = supervisions[order, 0]
                supervisions[:, -1] = input_lengths[order]
            else:
                supervisions[:, -1] = input_lengths
            self.__step_indices = None
            self.__supervisions_add = None
        return k2.DenseFsaVec(log_probs_eps, supervisions)

    def _maybe_normalize_gradients(self, log_probs: torch.Tensor, input_lengths: torch.Tensor) -> torch.Tensor:
        """Not required for RNNT.
        """
        return log_probs



# Copyright (c) 2021, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import torch
from einops import rearrange
from torch import nn

from nemo.collections.tts.modules.submodules import ConditionalInput, ConvNorm
from nemo.collections.tts.parts.utils.helpers import binarize_attention_parallel


class AlignmentEncoder(torch.nn.Module):
    """
    Module for alignment text and mel spectrogram.

    Args:
        n_mel_channels: Dimension of mel spectrogram.
        n_text_channels: Dimension of text embeddings.
        n_att_channels: Dimension of model
        temperature: Temperature to scale distance by.
            Suggested to be 0.0005 when using dist_type "l2" and 15.0 when using "cosine".
        condition_types: List of types for nemo.collections.tts.modules.submodules.ConditionalInput.
        dist_type: Distance type to use for similarity measurement. Supports "l2" and "cosine" distance.
    """

    def __init__(
        self,
        n_mel_channels=80,
        n_text_channels=512,
        n_att_channels=80,
        temperature=0.0005,
        condition_types=[],
        dist_type="l2",
    ):
        super().__init__()
        self.temperature = temperature
        self.cond_input = ConditionalInput(n_text_channels, n_text_channels, condition_types)
        self.softmax = torch.nn.Softmax(dim=3)
        self.log_softmax = torch.nn.LogSoftmax(dim=3)

        self.key_proj = nn.Sequential(
            ConvNorm(n_text_channels, n_text_channels * 2, kernel_size=3, bias=True, w_init_gain='relu'),
            torch.nn.ReLU(),
            ConvNorm(n_text_channels * 2, n_att_channels, kernel_size=1, bias=True),
        )

        self.query_proj = nn.Sequential(
            ConvNorm(n_mel_channels, n_mel_channels * 2, kernel_size=3, bias=True, w_init_gain='relu'),
            torch.nn.ReLU(),
            ConvNorm(n_mel_channels * 2, n_mel_channels, kernel_size=1, bias=True),
            torch.nn.ReLU(),
            ConvNorm(n_mel_channels, n_att_channels, kernel_size=1, bias=True),
        )
        if dist_type == "l2":
            self.dist_fn = self.get_euclidean_dist
        elif dist_type == "cosine":
            self.dist_fn = self.get_cosine_dist
        else:
            raise ValueError(f"Unknown distance type '{dist_type}'")

    @staticmethod
    def _apply_mask(inputs, mask, mask_value):
        if mask is None:
            return

        mask = rearrange(mask, "B T2 1 -> B 1 1 T2")
        inputs.data.masked_fill_(mask, mask_value)

    def get_dist(self, keys, queries, mask=None):
        """Calculation of distance matrix.

        Args:
            queries (torch.tensor): B x C1 x T1 tensor (probably going to be mel data).
            keys (torch.tensor): B x C2 x T2 tensor (text data).
            mask (torch.tensor): B x T2 x 1 tensor, binary mask for variable length entries and also can be used
                for ignoring unnecessary elements from keys in the resulting distance matrix (True = mask element, False = leave unchanged).
        Output:
            dist (torch.tensor): B x T1 x T2 tensor.
        """
        # B x C x T1
        queries_enc = self.query_proj(queries)
        # B x C x T2
        keys_enc = self.key_proj(keys)
        # B x 1 x T1 x T2
        dist = self.dist_fn(queries_enc=queries_enc, keys_enc=keys_enc)

        self._apply_mask(dist, mask, float("inf"))

        return dist.squeeze(1)

    @staticmethod
    def get_euclidean_dist(queries_enc, keys_enc):
        queries_enc = rearrange(queries_enc, "B C T1 -> B C T1 1")
        keys_enc = rearrange(keys_enc, "B C T2 -> B C 1 T2")
        # B x C x T1 x T2
        distance = (queries_enc - keys_enc) ** 2
        # B x 1 x T1 x T2
        l2_dist = distance.sum(axis=1, keepdim=True)
        return l2_dist

    @staticmethod
    def get_cosine_dist(queries_enc, keys_enc):
        queries_enc = rearrange(queries_enc, "B C T1 -> B C T1 1")
        keys_enc = rearrange(keys_enc, "B C T2 -> B C 1 T2")
        cosine_dist = -torch.nn.functional.cosine_similarity(queries_enc, keys_enc, dim=1)
        cosine_dist = rearrange(cosine_dist, "B T1 T2 -> B 1 T1 T2")
        return cosine_dist

    @staticmethod
    def get_durations(attn_soft, text_len, spect_len):
        """Calculation of durations.

        Args:
            attn_soft (torch.tensor): B x 1 x T1 x T2 tensor.
            text_len (torch.tensor): B tensor, lengths of text.
            spect_len (torch.tensor): B tensor, lengths of mel spectrogram.
        """
        attn_hard = binarize_attention_parallel(attn_soft, text_len, spect_len)
        durations = attn_hard.sum(2)[:, 0, :]
        assert torch.all(torch.eq(durations.sum(dim=1), spect_len))
        return durations

    @staticmethod
    def get_mean_dist_by_durations(dist, durations, mask=None):
        """Select elements from the distance matrix for the given durations and mask and return mean distance.

        Args:
            dist (torch.tensor): B x T1 x T2 tensor.
            durations (torch.tensor): B x T2 tensor. Dim T2 should sum to T1.
            mask (torch.tensor): B x T2 x 1 binary mask for variable length entries and also can be used
                for ignoring unnecessary elements in dist by T2 dim (True = mask element, False = leave unchanged).
        Output:
            mean_dist (torch.tensor): B x 1 tensor.
        """
        batch_size, t1_size, t2_size = dist.size()
        assert torch.all(torch.eq(durations.sum(dim=1), t1_size))

        AlignmentEncoder._apply_mask(dist, mask, 0)

        # TODO(oktai15): make it more efficient
        mean_dist_by_durations = []
        for dist_idx in range(batch_size):
            mean_dist_by_durations.append(
                torch.mean(
                    dist[
                        dist_idx,
                        torch.arange(t1_size),
                        torch.repeat_interleave(torch.arange(t2_size), repeats=durations[dist_idx]),
                    ]
                )
            )

        return torch.tensor(mean_dist_by_durations, dtype=dist.dtype, device=dist.device)

    @staticmethod
    def get_mean_distance_for_word(l2_dists, durs, start_token, num_tokens):
        """Calculates the mean distance between text and audio embeddings given a range of text tokens.

        Args:
            l2_dists (torch.tensor): L2 distance matrix from Aligner inference. T1 x T2 tensor.
            durs (torch.tensor): List of durations corresponding to each text token. T2 tensor. Should sum to T1.
            start_token (int): Index of the starting token for the word of interest.
            num_tokens (int): Length (in tokens) of the word of interest.
        Output:
            mean_dist_for_word (float): Mean embedding distance between the word indicated and its predicted audio frames.
        """
        # Need to calculate which audio frame we start on by summing all durations up to the start token's duration
        start_frame = torch.sum(durs[:start_token]).data

        total_frames = 0
        dist_sum = 0

        # Loop through each text token
        for token_ind in range(start_token, start_token + num_tokens):
            # Loop through each frame for the given text token
            for frame_ind in range(start_frame, start_frame + durs[token_ind]):
                # Recall that the L2 distance matrix is shape [spec_len, text_len]
                dist_sum += l2_dists[frame_ind, token_ind]

            # Update total frames so far & the starting frame for the next token
            total_frames += durs[token_ind]
            start_frame += durs[token_ind]

        return dist_sum / total_frames

    def forward(self, queries, keys, mask=None, attn_prior=None, conditioning=None):
        """Forward pass of the aligner encoder.

        Args:
            queries (torch.tensor): B x C1 x T1 tensor (probably going to be mel data).
            keys (torch.tensor): B x C2 x T2 tensor (text data).
            mask (torch.tensor): B x T2 x 1 tensor, binary mask for variable length entries (True = mask element, False = leave unchanged).
            attn_prior (torch.tensor): prior for attention matrix.
            conditioning (torch.tensor): B x 1 x C2 conditioning embedding
        Output:
            attn (torch.tensor): B x 1 x T1 x T2 attention mask. Final dim T2 should sum to 1.
            attn_logprob (torch.tensor): B x 1 x T1 x T2 log-prob attention mask.
        """
        keys = self.cond_input(keys.transpose(1, 2), conditioning).transpose(1, 2)
        # B x C x T1
        queries_enc = self.query_proj(queries)
        # B x C x T2
        keys_enc = self.key_proj(keys)
        # B x 1 x T1 x T2
        distance = self.dist_fn(queries_enc=queries_enc, keys_enc=keys_enc)
        attn = -self.temperature * distance

        if attn_prior is not None:
            attn = self.log_softmax(attn) + torch.log(attn_prior[:, None] + 1e-8)

        attn_logprob = attn.clone()

        self._apply_mask(attn, mask, -float("inf"))

        attn = self.softmax(attn)  # softmax along T2
        return attn, attn_logprob



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import abc
from abc import abstractmethod
from dataclasses import dataclass, field
from typing import Callable, cast

import torch
import torch.nn as nn

from nemo.collections.asr.parts.context_biasing.boosting_graph_batched import (
    BoostingTreeModelConfig,
    GPUBoostingTreeModel,
)
from nemo.collections.asr.parts.submodules.ngram_lm import NGramGPULanguageModel
from nemo.collections.common.tokenizers import TokenizerSpec
from nemo.core.utils.optional_libs import TRITON_AVAILABLE, triton_required
from nemo.utils import logging

if TRITON_AVAILABLE:
    import triton

    from nemo.collections.asr.parts.submodules.ngram_lm.ngram_lm_triton import ngram_multi_advance_triton_kernel

_BIASING_MODEL_CACHE = dict()


@dataclass
class BiasingRequestItemConfig:
    boosting_model_cfg: BoostingTreeModelConfig = field(default_factory=BoostingTreeModelConfig)
    boosting_model_alpha: float = 1.0  # boosting weight
    cache_key: str | None = None  # cache key for memory cache; NB: cache key should be unique for (tokenizer, phrases)
    multi_model_id: int | None = None  # compiled model id
    auto_manage_multi_model: bool = True  # if model should be added to the decoder and removed automatically

    def __post_init__(self):
        # if BiasingRequestItemConfig initialized from dict, we need to fix boosting_model_cfg field
        # see solution https://stackoverflow.com/a/60383031
        if isinstance(self.boosting_model_cfg, dict):
            self.boosting_model_cfg = BoostingTreeModelConfig(**self.boosting_model_cfg)

    def is_empty(self) -> bool:
        """Return True if biasing request (or model) is empty"""
        if self.cache_key and self.cache_key in _BIASING_MODEL_CACHE:
            return False
        if self.multi_model_id is not None:
            return False
        if not BoostingTreeModelConfig.is_empty(self.boosting_model_cfg):
            return False
        return True

    def get_model(self, tokenizer: TokenizerSpec) -> NGramGPULanguageModel | GPUBoostingTreeModel | None:
        """Create biasing model or get from cache, return the model. `None` is returned if biasing config is empty"""
        if self.cache_key and self.cache_key in _BIASING_MODEL_CACHE:
            return _BIASING_MODEL_CACHE[self.cache_key]
        if self.boosting_model_cfg.is_empty(self.boosting_model_cfg):
            return None
        boosting_model = GPUBoostingTreeModel.from_config(self.boosting_model_cfg, tokenizer=tokenizer)
        if self.cache_key:
            _BIASING_MODEL_CACHE[self.cache_key] = boosting_model
        return boosting_model

    def add_to_multi_model(self, tokenizer: TokenizerSpec, biasing_multi_model: "GPUBiasingMultiModelBase"):
        """Add biasing model to biasing multi-model"""
        boosting_model = self.get_model(tokenizer=tokenizer)
        if boosting_model is None:
            raise ValueError("Nothing to add, biasing model is empty")
        self.multi_model_id = biasing_multi_model.add_model(model=boosting_model, alpha=self.boosting_model_alpha)

    def remove_from_cache(self):
        """Remove model from cache (if cache entry exists)"""
        if self.cache_key and self.cache_key in _BIASING_MODEL_CACHE:
            del _BIASING_MODEL_CACHE[self.cache_key]

    def remove_from_multi_model(self, biasing_multi_model: "GPUBiasingMultiModelBase"):
        """Remove biasing model from multi-model"""
        if self.multi_model_id is None:
            # nothing to remove
            return
        biasing_multi_model.remove_model(self.multi_model_id)
        self.multi_model_id = None


class GPUBiasingMultiModelBase(abc.ABC, nn.Module):
    """
    Base class for implementing biasing multi-model:
    model that contains multiple biasing models and handles batched requests for them
    """

    START_STATE = 0

    @abstractmethod
    def add_model(self, model: NGramGPULanguageModel, alpha: float = 1.0) -> int:
        pass

    @abstractmethod
    def remove_model(self, model_id: int):
        pass

    @abstractmethod
    def has_models(self) -> bool:
        """Return True if the multi-model has at least one model"""
        pass

    def compatible_with_cuda_graphs(self) -> bool:
        """True if model can be compiled as a part of CUDA graph, False otherwise"""
        return False

    @abstractmethod
    def advance(
        self, states: torch.Tensor, model_ids: torch.Tensor, eos_id: int | None = None
    ) -> tuple[torch.Tensor, torch.Tensor]:
        """
        Advance `states` [B]: return scores [B, V] and next states [B, V] for full vocab
        Args:
            states: batch of states
            model_ids: ids of models for each state
            eos_id: if not None, for eos symbol use final state weight

        Returns:
            tuple with next states and scores
        """
        pass

    @abstractmethod
    def get_init_states(self, batch_size: int, bos=True) -> torch.Tensor:
        """
        Get batch of the initial states

        Args:
            batch_size: batch size
            bos: use begin-of-sentence state

        Returns:
            tensor [B] of initial states
        """
        pass


class GPUBiasingMultiModelReference(GPUBiasingMultiModelBase):
    """Reference implementation (incompatible with CUDA graphs)"""

    def __init__(self, vocab_size: int, *args, **kwargs):
        """

        Args:
            vocab_size: vocabulary size of the model
            *args, **kwargs: added for easiness of switching between this model and efficient implementation
        """
        super().__init__()
        self.models = nn.ModuleList([])
        self.buffer_for_device_handling = nn.Buffer(torch.zeros([1], dtype=torch.long))
        self.alphas: list[float] = []
        self.vocab_size: int = vocab_size
        self.float_dtype: torch.dtype | None = None
        self.bos_state: int | None = None
        self._params_defined = False
        self.free_ids = set()
        self.num_models = 0

    def has_models(self) -> bool:
        """Return True if the multi-model has at least one model"""
        return self.num_models > 0

    def _check_model_compatibility(self, model: NGramGPULanguageModel):
        if self.vocab_size != model.vocab_size:
            raise ValueError(f"Inconsistent vocab size: {model.vocab_size}")
        if self.bos_state != model.bos_state:
            raise ValueError(f"Inconsistent bos state: {self.bos_state} vs {model.bos_state}")
        if self.START_STATE != model.START_STATE:
            raise ValueError(f"Inconsistent start state: {self.START_STATE} vs {model.START_STATE}")

    def add_model(self, model: NGramGPULanguageModel, alpha: float = 1.0) -> int:
        if not self._params_defined:
            # there were no previous models
            self.bos_state = model.bos_state
            self.float_dtype = model.arcs_weights.dtype
            self._params_defined = True
        self._check_model_compatibility(model=model)
        try:
            model_id = self.free_ids.pop()
        except KeyError:
            model_id = None
        if model_id is None:
            model_id = len(self.models)
            self.models.append(model)
            self.alphas.append(alpha)
        else:
            self.models[model_id] = model
            self.alphas[model_id] = alpha
        self.num_models += 1
        return model_id

    def remove_model(self, model_id: int):
        self.models[model_id] = nn.Identity()  # dummy nn model
        self.alphas[model_id] = 0.0
        self.free_ids.add(model_id)
        self.num_models -= 1

    def get_init_states(self, batch_size: int, bos=True) -> torch.Tensor:
        """
        Get batch of the initial states

        Args:
            batch_size: batch size
            bos: use begin-of-sentence state

        Returns:
            tensor [B] of initial states
        """
        device = self.buffer_for_device_handling.device
        if not self._params_defined:
            return torch.zeros([batch_size], device=device, dtype=torch.long)
        return torch.full(
            [batch_size], fill_value=self.bos_state if bos else self.START_STATE, device=device, dtype=torch.long
        )

    def advance(
        self, states: torch.Tensor, model_ids: torch.Tensor, eos_id: int | None = None
    ) -> tuple[torch.Tensor, torch.Tensor]:
        """
        Advance `states` [B]: return scores [B, V] and next states [B, V] for full vocab
        Args:
            states: batch of states
            model_ids: ids of models for each state
            eos_id: if not None, for eos symbol use final state weight

        Returns:
            tuple with next states and scores
        """
        batch_size = states.shape[0]
        assert model_ids.shape[0] == batch_size
        device = next(iter(self.parameters())).device
        scores = torch.zeros([batch_size, self.vocab_size], device=device, dtype=self.float_dtype)
        next_states = torch.full([batch_size, self.vocab_size], fill_value=-1, dtype=torch.long, device=device)
        model_ids = model_ids.to("cpu").tolist()
        for batch_i, model_id in enumerate(model_ids):
            if model_id < 0:
                continue
            model = cast(NGramGPULanguageModel, self.models[model_id])
            scores_i, next_states_i = model.advance(states[batch_i : batch_i + 1], eos_id=eos_id)
            scores[batch_i : batch_i + 1] = scores_i * self.alphas[model_id]
            next_states[batch_i : batch_i + 1] = next_states_i
        return scores, next_states


class GPUBiasingMultiModel(GPUBiasingMultiModelBase):
    """Efficient multi-model implementation"""

    INIT_NUM_ARCS = 1_000_000
    INIT_NUM_STATES = 1_000_000
    INIT_NUM_MODELS = 128

    def __init__(
        self, vocab_size: int, reallocation_callback_fn: Callable | None = None, use_triton: bool | None = None
    ):
        """

        Args:
            vocab_size: vocabulary size of the model
            reallocation_callback_fn: function to call when reallocation occurred (needed for decoders with CUDA graphs)
            use_triton: allow using Triton, `None` means "auto" (used if available)
        """
        super().__init__()
        self.vocab_size: int = vocab_size
        self.float_dtype: torch.dtype | None = None
        self.bos_state: int | None = None
        self._params_defined = False
        self.free_ids = set()

        self.reallocation_callbacks = []
        if reallocation_callback_fn is not None:
            self.reallocation_callbacks.append(reallocation_callback_fn)

        self.use_triton = use_triton if use_triton is not None else TRITON_AVAILABLE

        int_dtype = torch.int64

        self.num_models = 0
        self.num_models_reserved = self.INIT_NUM_MODELS

        # store each model properties
        self.model2alpha = nn.Buffer(torch.zeros([self.num_models_reserved]))
        self.model2active = nn.Buffer(torch.zeros([self.num_models_reserved], dtype=torch.bool))
        self.model2num_states = nn.Buffer(torch.zeros([self.num_models_reserved], dtype=torch.int64))
        self.model2num_arcs = nn.Buffer(torch.zeros([self.num_models_reserved], dtype=torch.int64))
        self.model2num_arcs_extended = nn.Buffer(torch.zeros([self.num_models_reserved], dtype=torch.int64))
        self.model2states_offset = nn.Buffer(torch.zeros([self.num_models_reserved], dtype=torch.int64))
        self.model2arcs_offset = nn.Buffer(torch.zeros([self.num_models_reserved], dtype=torch.int64))

        self.num_states_total = 0
        self.num_arcs_extended_total = 0  # + extra padding
        self.num_states_reserved = self.INIT_NUM_STATES
        self.num_arcs_extended_reserved = self.INIT_NUM_ARCS  # + extra padding

        # arcs-related data
        self.all_arcs_weights = nn.Parameter(torch.zeros([self.num_arcs_extended_reserved]))
        self.all_from_states = nn.Buffer(torch.zeros([self.num_arcs_extended_reserved], dtype=int_dtype))
        self.all_to_states = nn.Buffer(torch.zeros([self.num_arcs_extended_reserved], dtype=int_dtype))
        self.all_ilabels = nn.Buffer(torch.zeros([self.num_arcs_extended_reserved], dtype=int_dtype))

        # states-related data
        self.all_start_end_arcs = nn.Buffer(torch.zeros([self.num_states_reserved, 2], dtype=int_dtype))
        self.all_state_order = nn.Buffer(torch.zeros([self.num_states_reserved], dtype=int_dtype))
        self.all_backoff_to_states = nn.Buffer(torch.zeros([self.num_states_reserved], dtype=int_dtype))
        self.all_backoff_weights = nn.Parameter(torch.zeros([self.num_states_reserved]))
        self.all_final_weights = nn.Parameter(torch.zeros([self.num_states_reserved]))

    def compatible_with_cuda_graphs(self) -> bool:
        """True if model can be compiled as a part of CUDA graph, False otherwise"""
        return self.use_triton

    def has_models(self) -> bool:
        """Return True if the multi-model has at least one model"""
        return self.num_models > 0

    def _check_model_compatibility(self, model: NGramGPULanguageModel):
        """Check that the new model parameters are the same compared to already stored models"""
        if self.vocab_size != model.vocab_size:
            raise ValueError(f"Inconsistent vocab size: {model.vocab_size}")
        if self.bos_state != model.bos_state:
            raise ValueError(f"Inconsistent bos state: {self.bos_state} vs {model.bos_state}")
        if self.START_STATE != model.START_STATE:
            raise ValueError(f"Inconsistent start state: {self.START_STATE} vs {model.START_STATE}")
        if not model._final_resolved:
            model._resolve_final()

    @staticmethod
    def _extend_buffer_or_param(buffer_or_param: nn.Buffer | nn.Parameter, add_len: int):
        """Extend buffer or parameter"""
        buffer_or_param.data = torch.cat(
            (
                buffer_or_param.data,
                torch.zeros(
                    [add_len] + list(buffer_or_param.shape)[1:],
                    dtype=buffer_or_param.dtype,
                    device=buffer_or_param.device,
                ),
            )
        )

    def _maybe_extend_arcs_and_states(self, add_num_states: int, add_num_arcs_extended: int) -> bool:
        """Extend memory allocated for arcs and states, return True if any tensor is reallocated"""
        reallocated = False

        if self.num_arcs_extended_total + add_num_arcs_extended > self.num_arcs_extended_reserved:
            # min allocation: 2x
            add_num_arcs = max(
                self.num_arcs_extended_reserved,
                self.num_arcs_extended_total + add_num_arcs_extended - self.num_arcs_extended_reserved,
            )
            self._extend_buffer_or_param(self.all_arcs_weights, add_len=add_num_arcs)
            self._extend_buffer_or_param(self.all_from_states, add_len=add_num_arcs)
            self._extend_buffer_or_param(self.all_to_states, add_len=add_num_arcs)
            self._extend_buffer_or_param(self.all_ilabels, add_len=add_num_arcs)
            self.num_arcs_extended_reserved += add_num_arcs
            reallocated = True

        if self.num_states_total + add_num_states > self.num_states_reserved:
            # min allocation: 2x
            add_num_states = max(
                self.num_states_reserved, self.num_states_total + add_num_states - self.num_states_reserved
            )
            self._extend_buffer_or_param(self.all_start_end_arcs, add_len=add_num_states)
            self._extend_buffer_or_param(self.all_state_order, add_len=add_num_states)
            self._extend_buffer_or_param(self.all_backoff_to_states, add_len=add_num_states)
            self._extend_buffer_or_param(self.all_backoff_weights, add_len=add_num_states)
            self._extend_buffer_or_param(self.all_final_weights, add_len=add_num_states)
            self.num_states_reserved += add_num_states
            reallocated = True

        return reallocated

    @staticmethod
    def _extend_buffer_2x(buffer: nn.Buffer):
        buffer.data = torch.cat((buffer.data, torch.zeros_like(buffer.data)), dim=-1)

    def _extend_num_models(self):
        """Extend memory allocated for models with properties"""
        assert self.num_models_reserved > 0
        self.num_models_reserved *= 2

        self._extend_buffer_2x(self.model2alpha)
        self._extend_buffer_2x(self.model2active)
        self._extend_buffer_2x(self.model2num_states)
        self._extend_buffer_2x(self.model2num_arcs)
        self._extend_buffer_2x(self.model2num_arcs_extended)
        self._extend_buffer_2x(self.model2states_offset)
        self._extend_buffer_2x(self.model2arcs_offset)

    @torch.no_grad()
    def add_model(self, model: GPUBoostingTreeModel, alpha: float = 1.0) -> int:
        """
        Add boosting model with `alpha` weight. Returns id for the added model

        Args:
            model: boosting model
            alpha: weight of the boosting model

        Returns:
            model id (to use in queries)
        """
        if not self._params_defined:
            # there were no previous models
            self.bos_state = model.bos_state
            self.float_dtype = model.arcs_weights.dtype
            self._params_defined = True
        self._check_model_compatibility(model=model)

        reallocated = False
        # select model id: either any free id, or num_models
        if self.free_ids:
            model_id = self.free_ids.pop()
        else:
            if self.num_models >= self.num_models_reserved:
                self._extend_num_models()
                reallocated = True
            model_id = self.num_models
            self.num_models += 1
        self.model2alpha[model_id] = alpha
        self.model2active[model_id] = True

        reallocated |= self._maybe_extend_arcs_and_states(
            add_num_states=model.num_states,
            add_num_arcs_extended=model.num_arcs_extended,
        )
        self.model2num_states[model_id] = model.num_states
        self.model2num_arcs[model_id] = model.num_arcs
        self.model2num_arcs_extended[model_id] = model.num_arcs_extended
        self.model2states_offset[model_id] = self.num_states_total
        self.model2arcs_offset[model_id] = self.num_arcs_extended_total

        # model is added always to the end of data storage
        states_start = self.num_states_total
        arcs_start = self.num_arcs_extended_total

        # arcs-related data
        self.all_arcs_weights.data[arcs_start : arcs_start + model.num_arcs].copy_(
            model.arcs_weights.data[: model.num_arcs]
        )
        self.all_from_states.data[arcs_start : arcs_start + model.num_arcs].copy_(
            model.from_states.data[: model.num_arcs]
        )
        self.all_to_states.data[arcs_start : arcs_start + model.num_arcs].copy_(model.to_states.data[: model.num_arcs])
        self.all_ilabels.data[arcs_start : arcs_start + model.num_arcs].copy_(model.ilabels.data[: model.num_arcs])

        # states-related data
        self.all_start_end_arcs.data[states_start : states_start + model.num_states].copy_(
            model.start_end_arcs.data[: model.num_states]
        )
        self.all_state_order.data[states_start : states_start + model.num_states].copy_(
            model.state_order.data[: model.num_states]
        )
        self.all_backoff_to_states.data[states_start : states_start + model.num_states].copy_(
            model.backoff_to_states.data[: model.num_states]
        )
        self.all_backoff_weights.data[states_start : states_start + model.num_states].copy_(
            model.backoff_weights.data[: model.num_states]
        )
        self.all_final_weights.data[states_start : states_start + model.num_states].copy_(
            model.final_weights.data[: model.num_states]
        )

        self.num_states_total += model.num_states
        self.num_arcs_extended_total += model.num_arcs_extended

        if reallocated:
            logging.info("Biasing multi-model reallocated memory. Executing reallocation callbacks")
            for reallocation_callback_fn in self.reallocation_callbacks:
                reallocation_callback_fn()
        return model_id

    @staticmethod
    def _clear_buffer_or_param_range(
        buffer_or_param: nn.Buffer | nn.Parameter, start: int, end: int, buffer_len: int | None = None
    ):
        if buffer_len is None:
            buffer_len = buffer_or_param.shape[0]
        remove_len = end - start
        buffer_or_param[start : buffer_len - remove_len].copy_(buffer_or_param[end:buffer_len].clone())
        buffer_or_param[buffer_len - remove_len : buffer_len].fill_(0)

    @torch.no_grad()
    def remove_model(self, model_id: int):
        """
        Remove boosting model.

        Args:
            model_id: boosting model id provided by the `add_model` method
        """
        logging.debug(f"Removing model: {model_id}")
        if model_id in self.free_ids or model_id >= self.num_models:
            raise ValueError(
                f"Trying to remove already deleted or non-existing model {model_id}. Total models in reserve: {self.num_models}"
            )

        # set model as inactive (we do not decrease num_models, only set to inactive!)
        self.model2active[model_id] = False
        self.model2alpha[model_id] = 0.0
        self.free_ids.add(model_id)

        start_state = self.model2states_offset[model_id].item()
        num_states = self.model2num_states[model_id].item()
        end_state = start_state + num_states

        start_arc = self.model2arcs_offset[model_id].item()
        num_arcs = self.model2num_arcs_extended[model_id].item()
        end_arc = start_arc + num_arcs

        assert num_arcs > 0 and num_states > 0, "Unexpected zero-size model"

        # clean up arcs-related data: cut [start_arc, end_arc) from the buffer (shifting right part to the left)
        self._clear_buffer_or_param_range(self.all_arcs_weights, start_arc, end_arc, self.num_arcs_extended_total)
        self._clear_buffer_or_param_range(self.all_from_states, start_arc, end_arc, self.num_arcs_extended_total)
        self._clear_buffer_or_param_range(self.all_to_states, start_arc, end_arc, self.num_arcs_extended_total)
        self._clear_buffer_or_param_range(self.all_ilabels, start_arc, end_arc, self.num_arcs_extended_total)

        # clean up states-related data: cut [start_state, end_state) from the buffer (shifting right part to the left)
        self._clear_buffer_or_param_range(self.all_start_end_arcs, start_state, end_state, self.num_states_total)
        self._clear_buffer_or_param_range(self.all_state_order, start_state, end_state, self.num_states_total)
        self._clear_buffer_or_param_range(self.all_backoff_to_states, start_state, end_state, self.num_states_total)
        self._clear_buffer_or_param_range(self.all_backoff_weights, start_state, end_state, self.num_states_total)
        self._clear_buffer_or_param_range(self.all_final_weights, start_state, end_state, self.num_states_total)

        # set num states/arcs to zero
        self.num_states_total -= num_states
        self.num_arcs_extended_total -= num_arcs

        self.model2num_states[model_id] = 0
        self.model2num_arcs[model_id] = 0
        self.model2num_arcs_extended[model_id] = 0
        # shift model offsets
        self.model2states_offset[model_id] = 0
        self.model2arcs_offset[model_id] = 0
        # shift states and arcs offsets
        torch.where(
            self.model2states_offset < start_state,
            self.model2states_offset,
            self.model2states_offset - num_states,
            out=self.model2states_offset,
        )
        torch.where(
            self.model2arcs_offset < start_arc,
            self.model2arcs_offset,
            self.model2arcs_offset - num_arcs,
            out=self.model2arcs_offset,
        )

    def get_init_states(self, batch_size: int, bos=True) -> torch.Tensor:
        """
        Get batch of the initial states

        Args:
            batch_size: batch size
            bos: use begin-of-sentence state

        Returns:
            tensor [B] of initial states
        """
        device = self.all_arcs_weights.device
        if not self._params_defined:
            return torch.full([batch_size], fill_value=self.START_STATE, device=device, dtype=torch.long)
        return torch.full(
            [batch_size], fill_value=self.bos_state if bos else self.START_STATE, device=device, dtype=torch.long
        )

    def advance(
        self, states: torch.Tensor, model_ids: torch.Tensor, eos_id: int | None = None
    ) -> tuple[torch.Tensor, torch.Tensor]:
        """
        Advance `states` [B]: return scores [B, V] and next states [B, V] for full vocab
        Args:
            states: batch of states
            model_ids: batch of ids of the models (`-1` to apply dummy model with zero weight)
            eos_id: if not None, for eos symbol use final state weight

        Returns:
            tuple with next states and scores
        """
        assert model_ids.shape[0] == states.shape[0]

        if self.use_triton and states.device.type == "cuda":
            scores, next_states = self._advance_triton(states=states, model_ids=model_ids)
        else:
            scores, next_states = self._advance_pytorch(states=states, model_ids=model_ids)
        # NB: model_id can be -1, but we assume that there at least 1 element in self.alphas
        scores *= self.model2alpha[model_ids][:, None]

        # replace eos_id score with maximum state weight to prevent from hallucinating in case of AED models (e.g. Canary)
        if eos_id is not None:
            raise NotImplementedError

        return scores, next_states

    @triton_required
    def _advance_triton(self, states: torch.Tensor, model_ids: torch.Tensor) -> tuple[torch.Tensor, torch.Tensor]:
        """
        Advance `states` [B]: return scores [B, V] and next states [B, V] for full vocab.
        Triton implementation. Currently not differentiable.

        Args:
            states: batch of states
            model_ids: ids of the models (`-1` to apply dummy model with zero weight)

        Returns:
            tuple of scores and next states
        """
        batch_size = states.shape[0]
        device = states.device
        scores = torch.zeros([batch_size, self.vocab_size], device=device, dtype=self.all_arcs_weights.dtype)
        next_states = torch.full([batch_size, self.vocab_size], fill_value=-1, dtype=torch.long, device=device)

        ngram_multi_advance_triton_kernel[batch_size,](
            vocab_size=self.vocab_size,
            states_ptr=states,
            new_states_out_ptr=next_states,
            scores_out_ptr=scores,
            start_state=self.START_STATE,
            model_ids_ptr=model_ids,
            states_offsets_ptr=self.model2states_offset,
            arcs_offsets_ptr=self.model2arcs_offset,
            to_states_ptr=self.all_to_states,
            ilabels_ptr=self.all_ilabels,
            arcs_weights_ptr=self.all_arcs_weights,
            start_end_arcs_ptr=self.all_start_end_arcs,
            backoff_to_states_ptr=self.all_backoff_to_states,
            backoff_weights_ptr=self.all_backoff_weights,
            BLOCK_SIZE=triton.next_power_of_2(self.vocab_size),
        )

        return scores, next_states

    def _advance_pytorch(self, states: torch.Tensor, model_ids: torch.Tensor) -> tuple[torch.Tensor, torch.Tensor]:
        """
        Advance `states` [B]: return scores [B, V] and next states [B, V] for full vocab.
        PyTorch implementation (slow, differentiable).

        Args:
            states: batch of states
            model_ids: ids of the models (`-1` to apply dummy model with zero weight)

        Returns:
            tuple of scores and next states
        """
        batch_size = states.shape[0]
        device = states.device
        current_states = states.clone()
        states_dtype = current_states.dtype

        # init output tensors
        out_scores = torch.zeros(batch_size, self.vocab_size, device=device)
        out_states = torch.full([batch_size, self.vocab_size], fill_value=-1, dtype=states_dtype, device=device)

        # helper ranges
        vocab_range = torch.arange(self.vocab_size, device=device)
        batch_indices = torch.arange(batch_size, device=device)

        # backoff weight accumulator
        accumulated_backoff = torch.zeros(batch_size, device=device)
        # loop condition
        start_state_not_processed = model_ids != -1

        states_offsets = self.model2states_offset[model_ids]
        arcs_offsets = self.model2arcs_offset[model_ids]

        num_iterations = 0
        while start_state_not_processed.any():
            num_iterations += 1
            # get arc boundaries
            start, end = self.all_start_end_arcs[current_states + states_offsets].unbind(dim=1)
            # number of arcs for each state cannot be larger than vocab size
            start += arcs_offsets
            end += arcs_offsets
            arc_indices = start[:, None] + vocab_range[None, :]
            mask = arc_indices < end[:, None]
            mask &= start_state_not_processed[:, None]
            mask_flat = mask.view(-1)
            arc_indices_flat = arc_indices.view(-1)
            # map indices outside the mask to vocab_size + 1
            scores_add = torch.zeros([batch_size, self.vocab_size + 1], device=device, dtype=out_scores.dtype)
            out_states_add = torch.full(
                [batch_size, self.vocab_size + 1], fill_value=-1, device=device, dtype=states_dtype
            )
            ilabels = self.all_ilabels[arc_indices_flat] * mask_flat + ~mask_flat * self.vocab_size
            scores_add[batch_indices.repeat_interleave(self.vocab_size), ilabels] = self.all_arcs_weights[
                arc_indices_flat
            ]
            out_states_add[batch_indices.repeat_interleave(self.vocab_size), ilabels] = self.all_to_states[
                arc_indices_flat
            ].to(states_dtype)
            # fill out_scores and out_states with new values where state is not found yet
            state_found = out_states != -1
            out_scores = torch.where(
                state_found, out_scores, accumulated_backoff.unsqueeze(-1) + scores_add[:, : self.vocab_size]
            )
            out_states = torch.where(state_found, out_states, out_states_add[:, : self.vocab_size])
            # update loop condition; process backoffs
            start_state_not_processed &= current_states != self.START_STATE
            accumulated_backoff += (
                self.all_backoff_weights[current_states + states_offsets] * start_state_not_processed
            )
            torch.where(
                start_state_not_processed,
                self.all_backoff_to_states[current_states + states_offsets],
                current_states,
                out=current_states,
            )
        return out_scores, out_states



# Copyright (c) 2025, NVIDIA CORPORATION & AFFILIATES.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# MIT License
#
# Copyright (c) 2023 Phil Wang
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import math
from functools import partial
from typing import Dict, Optional

import einops
import torch
import torch.nn.functional as F
from torch import nn
from torch.nn import Module

from nemo.core.classes import NeuralModule, typecheck
from nemo.core.neural_types import BoolType, FloatType, LengthsType, NeuralType, SpectrogramType
from nemo.utils import logging

__all__ = ['TransformerUNet']


class LearnedSinusoidalPosEmb(Module):
    """The sinusoidal Embedding to encode time conditional information"""

    def __init__(self, dim: int):
        super().__init__()
        if (dim % 2) != 0:
            raise ValueError(f"Input dimension {dim} is not divisible by 2!")
        half_dim = dim // 2
        self.weights = nn.Parameter(torch.randn(half_dim))

    def forward(self, t: torch.Tensor) -> torch.Tensor:
        """
        Args:
          t: input time tensor, shape (B)

        Return:
          fouriered: the encoded time conditional embedding, shape (B, D)
        """
        t = einops.rearrange(t, 'b -> b 1')
        freqs = t * einops.rearrange(self.weights, 'd -> 1 d') * 2 * math.pi
        fouriered = torch.cat((freqs.sin(), freqs.cos()), dim=-1)
        return fouriered


class ConvPositionEmbed(Module):
    """The Convolutional Embedding to encode time information of each frame"""

    def __init__(self, dim: int, kernel_size: int, groups: Optional[int] = None):
        super().__init__()
        if (kernel_size % 2) == 0:
            raise ValueError(f"Kernel size {kernel_size} is divisible by 2!")

        if groups is None:
            groups = dim

        self.dw_conv1d = nn.Sequential(
            nn.Conv1d(dim, dim, kernel_size, groups=groups, padding=kernel_size // 2), nn.GELU()
        )

    def forward(self, x, mask=None):
        """
        Args:
            x: input tensor, shape (B, T, D)

        Return:
            out: output tensor with the same shape (B, T, D)
        """

        if mask is not None:
            mask = mask[..., None]
            x = x.masked_fill(mask, 0.0)

        x = einops.rearrange(x, 'b n c -> b c n')
        x = self.dw_conv1d(x)
        out = einops.rearrange(x, 'b c n -> b n c')

        if mask is not None:
            out = out.masked_fill(mask, 0.0)

        return out


class RMSNorm(Module):
    """The Root Mean Square Layer Normalization

    References:
      - Zhang et al., Root Mean Square Layer Normalization, 2019
    """

    def __init__(self, dim):
        super().__init__()
        self.scale = dim**0.5
        self.gamma = nn.Parameter(torch.ones(dim))

    def forward(self, x: torch.Tensor):
        return F.normalize(x, dim=-1) * self.scale * self.gamma


class AdaptiveRMSNorm(Module):
    """
    Adaptive Root Mean Square Layer Normalization given a conditional embedding.
    This enables the model to consider the conditional input during normalization.
    """

    def __init__(self, dim: int, cond_dim: Optional[int] = None):
        super().__init__()
        if cond_dim is None:
            cond_dim = dim
        self.scale = dim**0.5

        self.to_gamma = nn.Linear(cond_dim, dim)
        self.to_beta = nn.Linear(cond_dim, dim)

        # init adaptive normalization to identity

        nn.init.zeros_(self.to_gamma.weight)
        nn.init.ones_(self.to_gamma.bias)

        nn.init.zeros_(self.to_beta.weight)
        nn.init.zeros_(self.to_beta.bias)

    def forward(self, x: torch.Tensor, cond: torch.Tensor):
        normed = F.normalize(x, dim=-1) * self.scale

        gamma, beta = self.to_gamma(cond), self.to_beta(cond)
        gamma = einops.rearrange(gamma, 'B D -> B 1 D')
        beta = einops.rearrange(beta, 'B D -> B 1 D')

        return normed * gamma + beta


class GEGLU(Module):
    """The GeGLU activation implementation"""

    def forward(self, x: torch.Tensor):
        x, gate = x.chunk(2, dim=-1)
        return F.gelu(gate) * x


def get_feedforward_layer(dim: int, mult: int = 4, dropout: float = 0.0):
    """
    Return a Feed-Forward layer for the Transformer Layer.
    GeGLU activation is used in this FF layer
    """
    dim_inner = int(dim * mult * 2 / 3)
    return nn.Sequential(nn.Linear(dim, dim_inner * 2), GEGLU(), nn.Dropout(dropout), nn.Linear(dim_inner, dim))


class TransformerUNet(NeuralModule):
    """
    Implementation of the transformer Encoder Model with U-Net structure used in
    VoiceBox and AudioBox

    References:
        Le et al., Voicebox: Text-Guided Multilingual Universal Speech Generation at Scale, 2023
        Vyas et al., Audiobox: Unified Audio Generation with Natural Language Prompts, 2023
    """

    def __init__(
        self,
        dim: int,
        depth: int,
        heads: int = 8,
        ff_mult: int = 4,
        attn_dropout: float = 0.0,
        ff_dropout: float = 0.0,
        max_positions: int = 6000,
        adaptive_rmsnorm: bool = False,
        adaptive_rmsnorm_cond_dim_in: Optional[int] = None,
        use_unet_skip_connection: bool = True,
        skip_connect_scale: Optional[int] = None,
    ):
        """
        Args:
            dim: Embedding dimension
            depth: Number of Transformer Encoder Layers
            heads: Number of heads in MHA
            ff_mult: The multiplier for the feedforward dimension (ff_dim = ff_mult * dim)
            attn_dropout: dropout rate for the MHA layer
            ff_dropout: droupout rate for the feedforward layer
            max_positions: The maximum time length of the input during training and inference
            adaptive_rmsnorm: Whether to use AdaptiveRMS layer.
                Set to True if the model has a conditional embedding in forward()
            adaptive_rms_cond_dim_in: Dimension of the conditional embedding
            use_unet_skip_connection: Whether to use U-Net or not
            skip_connect_scale: The scale of the U-Net connection.
        """
        super().__init__()
        if (depth % 2) != 0:
            raise ValueError(f"Number of layers {depth} is not divisible by 2!")
        self.layers = nn.ModuleList([])
        self.init_alibi(max_positions=max_positions, heads=heads)

        if adaptive_rmsnorm and adaptive_rmsnorm_cond_dim_in is None:
            raise ValueError("adaptive_rmsnorm_cond_dim_in must be provided if adaptive_rmsnorm is True")
        self.adaptive_rmsnorm = adaptive_rmsnorm
        self.adaptive_rmsnorm_cond_dim_in = adaptive_rmsnorm_cond_dim_in

        if self.adaptive_rmsnorm:
            rmsnorm_class = partial(AdaptiveRMSNorm, cond_dim=adaptive_rmsnorm_cond_dim_in)
        else:
            rmsnorm_class = RMSNorm

        if skip_connect_scale is None:
            self.skip_connect_scale = 2**-0.5
        else:
            self.skip_connect_scale = skip_connect_scale

        for ind in range(depth):
            layer = ind + 1
            has_skip = use_unet_skip_connection and layer > (depth // 2)

            self.layers.append(
                nn.ModuleList(
                    [
                        nn.Linear(dim * 2, dim) if has_skip else None,
                        rmsnorm_class(dim=dim),
                        nn.MultiheadAttention(
                            embed_dim=dim,
                            num_heads=heads,
                            dropout=attn_dropout,
                            batch_first=True,
                        ),
                        rmsnorm_class(dim=dim),
                        get_feedforward_layer(dim=dim, mult=ff_mult, dropout=ff_dropout),
                    ]
                )
            )

        self.final_norm = RMSNorm(dim)

        logging.debug('Initialized %s with', self.__class__.__name__)
        logging.debug('\tembedding dim:       %s', dim)
        logging.debug('\tNumber of Layer:     %s', depth)
        logging.debug('\tfeedforward dim:     %s', dim * ff_mult)
        logging.debug('\tnumber of heads:     %s', heads)
        logging.debug('\tDropout rate of MHA: %s', attn_dropout)
        logging.debug('\tDropout rate of FF:  %s', ff_dropout)
        logging.debug('\tnumber of heads:     %s', heads)
        logging.debug('\tmaximun time length: %s', max_positions)
        logging.debug('\tuse AdaptiveRMS:     %s', adaptive_rmsnorm)
        logging.debug('\tConditional  dim:    %s', adaptive_rmsnorm_cond_dim_in)
        logging.debug('\tUse UNet connection: %s', use_unet_skip_connection)
        logging.debug('\tskip connect scale:  %s', self.skip_connect_scale)

    def init_alibi(
        self,
        max_positions: int,
        heads: int,
    ):
        """Initialize the Alibi bias parameters

        References:
          - Press et al., Train Short, Test Long: Attention with Linear Biases Enables Input Length Extrapolation, 2021
        """

        def get_slopes(n):
            ratio = 2 ** (-8 / n)
            return ratio ** torch.arange(1, n + 1)

        if not math.log2(heads).is_integer():
            logging.warning(
                "It is recommend to set number of attention heads to be the power of 2 for the Alibi bias!"
            )
            logging.warning(f"Current value of heads: {heads}")

        self.slopes = nn.Parameter(einops.rearrange(get_slopes(heads), "B -> B 1 1"))

        pos_matrix = (
            -1 * torch.abs(torch.arange(max_positions).unsqueeze(0) - torch.arange(max_positions).unsqueeze(1)).float()
        )
        pos_matrix = einops.rearrange(pos_matrix, "T1 T2 -> 1 T1 T2")
        self.register_buffer('pos_matrix', pos_matrix, persistent=False)

    @property
    def input_types(self) -> Dict[str, NeuralType]:
        """Returns definitions of module output ports."""
        return {
            "x": NeuralType(('B', 'T', 'D'), FloatType()),
            "key_padding_mask": NeuralType(('B', 'T'), BoolType(), optional=True),
            "adaptive_rmsnorm_cond": NeuralType(('B', 'D'), FloatType(), optional=True),
        }

    @property
    def output_types(self) -> Dict[str, NeuralType]:
        """Returns definitions of module output ports."""
        return {
            "output": NeuralType(('B', 'T', 'D'), FloatType()),
        }

    @typecheck()
    def forward(self, x, key_padding_mask: Optional[torch.Tensor] = None, adaptive_rmsnorm_cond=None):
        """Forward pass of the model.

        Args:
            input: input tensor, shape (B, C, D, T)
            key_padding_mask: mask tensor indicating the padding parts, shape (B, T)
            adaptive_rmsnorm_cond: conditional input for the model, shape (B, D)
        """
        batch_size, seq_len, *_ = x.shape
        skip_connects = []
        alibi_bias = self.get_alibi_bias(batch_size=batch_size, seq_len=seq_len)

        rmsnorm_kwargs = dict()
        if adaptive_rmsnorm_cond is not None:
            rmsnorm_kwargs = dict(cond=adaptive_rmsnorm_cond)

        for skip_combiner, attn_prenorm, attn, ff_prenorm, ff in self.layers:

            if skip_combiner is None:
                skip_connects.append(x)
            else:
                skip_connect = skip_connects.pop() * self.skip_connect_scale
                x = torch.cat((x, skip_connect), dim=-1)
                x = skip_combiner(x)

            attn_input = attn_prenorm(x, **rmsnorm_kwargs)
            if key_padding_mask is not None:
                # Since Alibi_bias is a float-type attn_mask, the padding_mask need to be float-type.
                float_key_padding_mask = key_padding_mask.float()
                float_key_padding_mask = float_key_padding_mask.masked_fill(key_padding_mask, float('-inf'))
            else:
                float_key_padding_mask = None

            attn_output, _ = attn(
                query=attn_input,
                key=attn_input,
                value=attn_input,
                key_padding_mask=float_key_padding_mask,
                need_weights=False,
                attn_mask=alibi_bias,
            )
            x = x + attn_output

            ff_input = ff_prenorm(x, **rmsnorm_kwargs)
            x = ff(ff_input) + x

        return self.final_norm(x)

    def get_alibi_bias(self, batch_size: int, seq_len: int):
        """
        Return the alibi_bias given batch size and seqence length
        """
        pos_matrix = self.pos_matrix[:, :seq_len, :seq_len]
        alibi_bias = pos_matrix * self.slopes
        alibi_bias = alibi_bias.repeat(batch_size, 1, 1)

        return alibi_bias


class SpectrogramTransformerUNet(NeuralModule):
    """This model handles complex-valued inputs by stacking real and imaginary components.
    Stacked tensor is processed using TransformerUNet and the output is projected to generate real
    and imaginary components of the output channels.

    Convolutional Positional Embedding is applied for the input sequence
    """

    def __init__(
        self,
        in_channels: int = 1,
        out_channels: int = 1,
        freq_dim: int = 256,
        dim: int = 1024,
        depth: int = 24,
        heads: int = 16,
        ff_mult: int = 4,
        ff_dropout: float = 0.0,
        attn_dropout: float = 0.0,
        max_positions: int = 6000,
        time_hidden_dim: Optional[int] = None,
        conv_pos_embed_kernel_size: int = 31,
        conv_pos_embed_groups: Optional[int] = None,
        adaptive_rmsnorm: Optional[bool] = True,
    ):
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        dim_in = freq_dim * in_channels * 2

        if time_hidden_dim is None:
            time_hidden_dim = dim * 4

        self.proj_in = nn.Linear(dim_in, dim)
        if adaptive_rmsnorm:
            self.sinu_pos_emb = nn.Sequential(LearnedSinusoidalPosEmb(dim), nn.Linear(dim, time_hidden_dim), nn.SiLU())

        self.conv_embed = ConvPositionEmbed(
            dim=dim, kernel_size=conv_pos_embed_kernel_size, groups=conv_pos_embed_groups
        )

        self.transformerunet = TransformerUNet(
            dim=dim,
            depth=depth,
            heads=heads,
            ff_mult=ff_mult,
            ff_dropout=ff_dropout,
            attn_dropout=attn_dropout,
            max_positions=max_positions,
            adaptive_rmsnorm=adaptive_rmsnorm,
            adaptive_rmsnorm_cond_dim_in=time_hidden_dim,
            use_unet_skip_connection=True,
        )

        # 2x the frequency dimension as the model operates in the complex-value domain
        dim_out = freq_dim * out_channels * 2

        self.proj_out = nn.Linear(dim, dim_out)

        logging.debug('Initialized %s with', self.__class__.__name__)
        logging.debug('\tin_channels:  %s', self.in_channels)
        logging.debug('\tout_channels: %s', self.out_channels)
        logging.debug('\tInput frequency dimension: %s', freq_dim)

    @property
    def input_types(self) -> Dict[str, NeuralType]:
        """Returns definitions of module output ports."""
        return {
            "input": NeuralType(('B', 'C', 'D', 'T'), SpectrogramType()),
            "input_length": NeuralType(('B',), LengthsType(), optional=True),
            "condition": NeuralType(('B',), FloatType(), optional=True),
        }

    @property
    def output_types(self) -> Dict[str, NeuralType]:
        """Returns definitions of module output ports."""
        return {
            "output": NeuralType(('B', 'C', 'D', 'T'), SpectrogramType()),
            "output_length": NeuralType(('B',), LengthsType(), optional=True),
        }

    @staticmethod
    def _get_key_padding_mask(input_length: torch.Tensor, max_length: int):
        """
        Return the self_attention masking according to the input length.
        0 indicates the frame is in the valid range, while 1 indicates the frame is a padding frame.
        Args:
          input_length: shape (B)
          max_length (int): The maximum length of the input sequence

        return:
          key_padding_mask: shape (B, T)
        """
        key_padding_mask = torch.arange(max_length).expand(len(input_length), max_length).to(input_length.device)
        key_padding_mask = key_padding_mask >= input_length.unsqueeze(1)
        return key_padding_mask

    @typecheck()
    def forward(self, input, input_length=None, condition=None):
        """Forward pass of the model.

        Args:
            input: input tensor, shape (B, C, D, T)
            input_length: length of the valid time steps for each example in the batch, shape (B,)
            condition: scalar condition (time) for the model, will be embedded using `self.time_embedding`
        """
        # Stack real and imaginary components
        B, C_in, D, T = input.shape
        if C_in != self.in_channels:
            raise RuntimeError(f'Unexpected input channel size {C_in}, expected {self.in_channels}')

        input_real_imag = torch.stack([input.real, input.imag], dim=2)
        input = einops.rearrange(input_real_imag, 'B C RI D T -> B T (C RI D)')

        x = self.proj_in(input)
        key_padding_mask = self._get_key_padding_mask(input_length, max_length=T)
        x = self.conv_embed(x, mask=key_padding_mask) + x

        if condition is None:
            time_emb = None
        else:
            time_emb = self.sinu_pos_emb(condition)

        x = self.transformerunet(x=x, key_padding_mask=key_padding_mask, adaptive_rmsnorm_cond=time_emb)

        output = self.proj_out(x)
        output = einops.rearrange(output, "B T (C RI D) -> B C D T RI", C=self.out_channels, RI=2, D=D)
        output = torch.view_as_complex(output.contiguous())

        return output, input_length



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
from datetime import datetime
from typing import AsyncGenerator, List, Optional

from loguru import logger
from pipecat.frames.frames import (
    CancelFrame,
    EndFrame,
    ErrorFrame,
    Frame,
    InterimTranscriptionFrame,
    StartFrame,
    TranscriptionFrame,
    VADUserStartedSpeakingFrame,
    VADUserStoppedSpeakingFrame,
)
from pipecat.processors.frame_processor import FrameDirection
from pipecat.services.stt_service import STTService
from pipecat.transcriptions.language import Language
from pipecat.utils.time import time_now_iso8601
from pipecat.utils.tracing.service_decorators import traced_stt
from pydantic import BaseModel

from nemo.agents.voice_agent.pipecat.services.nemo.audio_logger import AudioLogger
from nemo.agents.voice_agent.pipecat.services.nemo.streaming_asr import NemoStreamingASRService

ASR_EOU_MODELS = ["nvidia/parakeet_realtime_eou_120m-v1"]

try:
    # disable nemo logging
    from nemo.utils import logging

    level = logging.getEffectiveLevel()
    logging.setLevel(logging.CRITICAL)


except ModuleNotFoundError as e:
    logger.error(f"Exception: {e}")
    logger.error('In order to use NVIDIA NeMo STT, you need to `pip install "nemo_toolkit[all]"`.')
    raise Exception(f"Missing module: {e}")


class NeMoSTTInputParams(BaseModel):
    """Input parameters for NeMo STT service."""

    language: Optional[Language] = Language.EN_US
    att_context_size: Optional[List] = [70, 1]
    frame_len_in_secs: Optional[float] = 0.08  # 80ms for FastConformer model
    config_path: Optional[str] = None  # path to the Niva ASR config file
    raw_audio_frame_len_in_secs: Optional[float] = 0.016  # 16ms for websocket transport
    buffer_size: Optional[int] = 5  # number of audio frames to buffer, 1 frame is 16ms


class NemoSTTService(STTService):
    """NeMo Speech-to-Text service for Pipecat integration."""

    def __init__(
        self,
        *,
        model: Optional[str] = "nvidia/stt_en_fastconformer_hybrid_large_streaming_multi",
        device: Optional[str] = "cuda:0",
        sample_rate: Optional[int] = 16000,
        params: Optional[NeMoSTTInputParams] = None,
        has_turn_taking: Optional[bool] = None,  # if None, it will be set by the model name
        backend: Optional[str] = "legacy",
        decoder_type: Optional[str] = "rnnt",
        audio_logger: Optional[AudioLogger] = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self._queue = asyncio.Queue()
        self._sample_rate = sample_rate
        params.buffer_size = params.frame_len_in_secs // params.raw_audio_frame_len_in_secs
        self._params = params
        self._model_name = model
        if has_turn_taking is None:
            has_turn_taking = True if model in ASR_EOU_MODELS else False
            logger.info(f"Setting has_turn_taking to `{has_turn_taking}` based on model name: `{model}`")
        self._has_turn_taking = has_turn_taking
        self._backend = backend
        self._decoder_type = decoder_type
        self._audio_logger = audio_logger
        self._is_vad_active = False
        if not params:
            raise ValueError("params is required")

        self._device = device

        self._load_model()

        self.audio_buffer = []
        self.user_is_speaking = False

    def _load_model(self):
        if self._backend == "legacy":
            self._model = NemoStreamingASRService(
                self._model_name,
                self._params.att_context_size,
                device=self._device,
                decoder_type=self._decoder_type,
                frame_len_in_secs=self._params.frame_len_in_secs,
            )
        else:
            raise ValueError(f"Invalid ASR backend: {self._backend}")

    def can_generate_metrics(self) -> bool:
        """Indicates whether this service can generate metrics.

        Returns:
            bool: True, as this service supports metric generation.
        """
        return True

    async def start(self, frame: StartFrame):
        """Handle service start.

        Args:
            frame: StartFrame containing initial configuration
        """
        await super().start(frame)

        # Initialize the model if not already done
        if not hasattr(self, "_model"):
            self._load_model()

    async def stop(self, frame: EndFrame):
        """Handle service stop.

        Args:
            frame: EndFrame that triggered this method
        """
        await super().stop(frame)
        # Clear any internal state if needed
        await self._queue.put(None)  # Signal to stop processing

    async def cancel(self, frame: CancelFrame):
        """Handle service cancellation.

        Args:
            frame: CancelFrame that triggered this method
        """
        await super().cancel(frame)
        # Clear any internal state
        await self._queue.put(None)  # Signal to stop processing
        self._queue = asyncio.Queue()  # Reset the queue

    async def run_stt(self, audio: bytes) -> AsyncGenerator[Frame, None]:
        """Process audio data and generate transcription frames.

        Args:
            audio: Raw audio bytes to transcribe

        Yields:
            Frame: Transcription frames containing the results
        """
        timestamp_now = datetime.now()
        await self.start_ttfb_metrics()
        await self.start_processing_metrics()
        if self._audio_logger is not None and self._audio_logger.first_audio_timestamp is None:
            self._audio_logger.first_audio_timestamp = timestamp_now

        try:
            is_final = False
            user_has_finished = False
            transcription = None
            self.audio_buffer.append(audio)
            if len(self.audio_buffer) >= self._params.buffer_size:
                audio = b"".join(self.audio_buffer)
                self.audio_buffer = []

                # Append to continuous user audio buffer for stereo conversation recording
                if self._audio_logger is not None:
                    self._audio_logger.append_continuous_user_audio(audio)

                asr_result = self._model.transcribe(audio)
                transcription = asr_result.text
                is_final = asr_result.is_final
                if self._audio_logger is not None:
                    if self._is_vad_active:
                        is_first_frame = False
                        self._audio_logger.turn_audio_buffer.append(audio)
                        # Accumulate transcriptions for turn-based logging
                        if transcription:
                            self._audio_logger.turn_transcription_buffer.append(transcription)
                            self._audio_logger.stage_turn_audio_and_transcription(
                                timestamp_now=timestamp_now,
                                is_first_frame=is_first_frame,
                                additional_metadata={
                                    "model": self._model_name,
                                    "backend": self._backend,
                                },
                            )
                eou_latency = asr_result.eou_latency
                eob_latency = asr_result.eob_latency
                eou_prob = asr_result.eou_prob
                eob_prob = asr_result.eob_prob
                if eou_latency is not None:
                    logger.debug(
                        f"EOU latency: {eou_latency: .4f} seconds. EOU probability: {eou_prob: .2f}."
                        f"Processing time: {asr_result.processing_time: .4f} seconds."
                    )
                    user_has_finished = True
                if eob_latency is not None:
                    logger.debug(
                        f"EOB latency: {eob_latency: .4f} seconds. EOB probability: {eob_prob: .2f}."
                        f"Processing time: {asr_result.processing_time: .4f} seconds."
                    )
                    user_has_finished = True
                await self.stop_ttfb_metrics()
                await self.stop_processing_metrics()

            if transcription:
                logger.debug(f"Transcription (is_final={is_final}): `{transcription}`")
                self.user_is_speaking = True if not user_has_finished else False

                # Get the language from params or default to EN_US
                language = self._params.language if self._params else Language.EN_US

                # Create and push the transcription frame
                if self._has_turn_taking:
                    # if turn taking is enabled, we push interim transcription frames
                    # and let the turn taking service handle the final transcription
                    frame_type = InterimTranscriptionFrame
                else:
                    # otherwise, we use the is_final flag to determine the frame type
                    frame_type = TranscriptionFrame if not is_final else InterimTranscriptionFrame
                await self.push_frame(
                    frame_type(
                        transcription,
                        "",  # No speaker ID in this implementation
                        time_now_iso8601(),
                        language,
                        result={"text": transcription},
                    )
                )

                # Handle the transcription
                await self._handle_transcription(
                    transcript=transcription,
                    is_final=is_final,
                    language=language,
                )

            yield None

        except Exception as e:
            logger.error(f"Error in NeMo STT processing: {e}")
            await self.push_frame(
                ErrorFrame(
                    str(e),
                    time_now_iso8601(),
                )
            )
            yield None

    @traced_stt
    async def _handle_transcription(self, transcript: str, is_final: bool, language: Optional[str] = None):
        """Handle a transcription result.

        Args:
            transcript: The transcribed text
            is_final: Whether this is a final transcription
            language: The language of the transcription
        """
        pass  # Base implementation - can be extended for specific handling needs

    async def set_language(self, language: Language):
        """Update the service's recognition language.

        Args:
            language: New language for recognition
        """
        if self._params:
            self._params.language = language
        else:
            self._params = NeMoSTTInputParams(language=language)

        logger.info(f"Switching STT language to: {language}")

    async def set_model(self, model: str):
        """Update the service's model.

        Args:
            model: New model name/path to use
        """
        await super().set_model(model)
        self._model_name = model
        self._load_model()

    async def process_frame(self, frame: Frame, direction: FrameDirection):
        """Process incoming frames and handle VAD events."""
        if isinstance(frame, VADUserStoppedSpeakingFrame) and isinstance(self._model, NemoStreamingASRService):
            # manualy reset the state of the model when end of utterance is detected by VAD
            logger.debug("Resetting state of the model due to VADUserStoppedSpeakingFrame")
            if self.user_is_speaking:
                logger.debug(
                    "[EOU missing] STT failed to detect end of utterance before VAD detected user stopped speaking"
                )
            self._model.reset_state()
            self._is_vad_active = False
        elif isinstance(frame, VADUserStartedSpeakingFrame):
            self._is_vad_active = True

        await super().process_frame(frame, direction)



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import nemo_run as run
import pytest
import torch
from megatron.core.distributed import DistributedDataParallelConfig

from nemo.collections.llm.api import finetune, pretrain
from nemo.collections.llm.gpt.data.mock import MockDataModule
from nemo.collections.llm.gpt.data.squad import SquadDataModule
from nemo.collections.llm.gpt.model.mixtral import MixtralConfig8x22B, MixtralModel
from nemo.collections.llm.peft.lora import LoRA
from nemo.collections.llm.recipes import mixtral_8x22b
from nemo.lightning import Trainer


class TestMixtral8x22B:
    @pytest.fixture(scope="class")
    def recipe_module(self):
        return mixtral_8x22b

    def test_model(self, recipe_module):
        model_config = recipe_module.model()
        assert isinstance(model_config, run.Config)
        assert model_config.__fn_or_cls__ == MixtralModel
        assert isinstance(model_config.config, run.Config)
        assert model_config.config.__fn_or_cls__ == MixtralConfig8x22B

    def test_trainer(self, recipe_module):
        trainer_config = recipe_module.trainer()
        assert isinstance(trainer_config, run.Config)
        assert trainer_config.__fn_or_cls__ == Trainer
        assert trainer_config.accelerator == "gpu"
        assert trainer_config.devices == 8
        assert trainer_config.num_nodes == 16

        # Check strategy configuration
        assert isinstance(trainer_config.strategy, run.Config)
        assert trainer_config.strategy.__fn_or_cls__.__name__ == "MegatronStrategy"
        assert trainer_config.strategy.tensor_model_parallel_size == 2
        assert trainer_config.strategy.pipeline_model_parallel_size == 4
        assert trainer_config.strategy.pipeline_dtype == torch.bfloat16
        assert trainer_config.strategy.virtual_pipeline_model_parallel_size == 14
        assert trainer_config.strategy.context_parallel_size == 2
        assert trainer_config.strategy.sequence_parallel is True
        assert trainer_config.strategy.expert_model_parallel_size == 8

        # Check DDP configuration
        assert isinstance(trainer_config.strategy.ddp, run.Config)
        assert trainer_config.strategy.ddp.__fn_or_cls__ == DistributedDataParallelConfig
        assert trainer_config.strategy.ddp.check_for_nan_in_grad is True
        assert trainer_config.strategy.ddp.grad_reduce_in_fp32 is True

    def test_pretrain_recipe(self, recipe_module):
        recipe = recipe_module.pretrain_recipe()
        assert isinstance(recipe, run.Partial)
        assert recipe.__fn_or_cls__ == pretrain
        assert isinstance(recipe.model, run.Config)
        assert recipe.model.__fn_or_cls__ == MixtralModel
        assert isinstance(recipe.trainer, run.Config)
        assert recipe.trainer.__fn_or_cls__ == Trainer
        assert isinstance(recipe.data, run.Config)
        assert recipe.data.__fn_or_cls__ == MockDataModule
        assert isinstance(recipe.model.config, run.Config)
        assert recipe.data.seq_length == 4096
        assert recipe.data.global_batch_size == 512
        assert recipe.data.micro_batch_size == 1

    def test_finetune_recipe(self, recipe_module):
        recipe = recipe_module.finetune_recipe()
        assert isinstance(recipe, run.Partial)
        assert recipe.__fn_or_cls__ == finetune
        assert isinstance(recipe.model, run.Config)
        assert recipe.model.__fn_or_cls__ == MixtralModel
        assert isinstance(recipe.trainer, run.Config)
        assert recipe.trainer.__fn_or_cls__ == Trainer
        assert isinstance(recipe.data, run.Config)
        assert recipe.data.__fn_or_cls__ == SquadDataModule
        assert recipe.data.seq_length == 2048
        assert recipe.data.global_batch_size == 128
        assert recipe.data.micro_batch_size == 1
        assert isinstance(recipe.peft, run.Config)
        assert recipe.peft.__fn_or_cls__ == LoRA
        assert recipe.peft.target_modules == ['linear_qkv', 'linear_proj']
        assert recipe.peft.dim == 32

    @pytest.mark.parametrize("num_nodes,num_gpus_per_node", [(8, 8), (16, 4), (32, 2)])
    def test_pretrain_recipe_with_different_configurations(self, recipe_module, num_nodes, num_gpus_per_node):
        recipe = recipe_module.pretrain_recipe(num_nodes=num_nodes, num_gpus_per_node=num_gpus_per_node)
        assert recipe.trainer.num_nodes == num_nodes
        assert recipe.trainer.devices == num_gpus_per_node

    def test_trainer_parallelism_options(self, recipe_module):
        trainer_config = recipe_module.trainer(
            tensor_parallelism=4,
            pipeline_parallelism=4,
            context_parallelism=2,
            sequence_parallelism=False,
            expert_parallelism=2,
        )
        assert trainer_config.strategy.tensor_model_parallel_size == 4
        assert trainer_config.strategy.pipeline_model_parallel_size == 4
        assert trainer_config.strategy.context_parallel_size == 2
        assert trainer_config.strategy.sequence_parallel is False
        assert trainer_config.strategy.expert_model_parallel_size == 2

    def test_model_config_parameters(self, recipe_module):
        model_config = recipe_module.model()
        mixtral_config = model_config.config
        assert isinstance(mixtral_config, run.Config)
        assert mixtral_config.num_layers == 56
        assert mixtral_config.hidden_size == 6144
        assert mixtral_config.num_attention_heads == 48
        assert mixtral_config.seq_length == 4096
        assert mixtral_config.num_moe_experts == 8



# Copyright (c) 2022, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import torch
from omegaconf import DictConfig

from nemo.collections.asr.models import EncDecDenoiseMaskedTokenPredModel, SpeechEncDecSelfSupervisedModel
from nemo.core.classes.common import typecheck


@pytest.fixture()
def ssl_model():
    preprocessor = {
        'cls': 'nemo.collections.asr.modules.AudioToMelSpectrogramPreprocessor',
        'params': dict({'pad_to': 16, 'dither': 0}),
    }

    model_defaults = {'enc_hidden': 32, 'dec_out': 128}

    encoder = {
        'cls': 'nemo.collections.asr.modules.ConvASREncoder',
        'params': {
            'feat_in': 64,
            'activation': 'relu',
            'conv_mask': True,
            'jasper': [
                {
                    'filters': model_defaults['enc_hidden'],
                    'repeat': 1,
                    'kernel': [1],
                    'stride': [1],
                    'dilation': [1],
                    'dropout': 0.0,
                    'residual': False,
                    'separable': True,
                    'se': True,
                    'se_context_size': -1,
                },
                {
                    'filters': model_defaults['enc_hidden'],
                    'repeat': 1,
                    'kernel': [1],
                    'stride': [1],
                    'dilation': [1],
                    'dropout': 0.0,
                    'residual': False,
                    'separable': True,
                    'se': True,
                    'se_context_size': -1,
                },
                {
                    'filters': model_defaults['enc_hidden'],
                    'repeat': 1,
                    'kernel': [1],
                    'stride': [1],
                    'dilation': [1],
                    'dropout': 0.0,
                    'residual': False,
                    'separable': True,
                    'se': True,
                    'se_context_size': -1,
                },
            ],
        },
    }

    spec_augment = {
        '_target_': 'nemo.collections.asr.modules.MaskedPatchAugmentation',
        'freq_masks': 3,
        'freq_width': 20,
        'patch_size': 16,
        'mask_patches': 0.5,
    }

    loss_list_contr_mlm = {
        'contr': {
            'decoder': {
                '_target_': 'nemo.collections.asr.modules.ConvASRDecoderReconstruction',
                'feat_in': model_defaults['enc_hidden'],
                'feat_hidden': 128,
                'feat_out': model_defaults['dec_out'],
                'stride_layers': 0,
                'non_stride_layers': 0,
                'stride_transpose': False,
            },
            'loss': {
                '_target_': 'nemo.collections.asr.losses.ContrastiveLoss',
                'in_dim': 64,
                'proj_dim': model_defaults['dec_out'],
                'combine_time_steps': 1,
                'quantized_targets': True,
                'codebook_size': 64,
                'sample_from_same_utterance_only': True,
                'sample_from_non_masked': False,
                'num_negatives': 3,
            },
        },
        'mlm': {
            'decoder': {
                '_target_': 'nemo.collections.asr.modules.ConvASRDecoder',
                'feat_in': model_defaults['enc_hidden'],
                'num_classes': 4096,
            },
            'loss': {'_target_': 'nemo.collections.asr.losses.MLMLoss', 'combine_time_steps': 1},
            'targets_from_loss': "contr",
        },
    }

    modelConfig_contr_mlm = DictConfig(
        {
            'preprocessor': DictConfig(preprocessor),
            'spec_augment': DictConfig(spec_augment),
            'model_defaults': DictConfig(model_defaults),
            'encoder': DictConfig(encoder),
            'loss_list': DictConfig(loss_list_contr_mlm),
        }
    )
    ssl_model = SpeechEncDecSelfSupervisedModel(cfg=modelConfig_contr_mlm)
    return ssl_model


@pytest.fixture()
def denoise_mlm_ssl_model():

    model_defaults = {
        "subsampling_factor": 1,
        'enc_hidden': 32,
        'dec_out': 128,
        "sample_rate": 16000,
        "num_classes": 32,
        "num_books": 1,
        "code_dim": 16,
        "squeeze_single": False,
        "mask_position": "pre_conv",  # position to apply masking, before or after conv subsampling, choices in ['pre_conv', 'post_conv']
    }

    preprocessor = {
        "_target_": "nemo.collections.asr.modules.AudioToMelSpectrogramPreprocessor",
        "sample_rate": model_defaults["sample_rate"],
        "normalize": "per_feature",
        "window_size": 0.025,
        "window_stride": 0.01,
        "window": "hann",
        "features": 80,
        "n_fft": 512,
        "log": True,
        "frame_splicing": 1,
        "dither": 0.00001,
        "pad_to": 16,
        "pad_value": 0.0,
    }

    encoder = {
        'cls': 'nemo.collections.asr.modules.ConvASREncoder',
        'params': {
            'feat_in': preprocessor["features"],
            'activation': 'relu',
            'conv_mask': True,
            'jasper': [
                {
                    'filters': model_defaults['enc_hidden'],
                    'repeat': 1,
                    'kernel': [1],
                    'stride': [1],
                    'dilation': [1],
                    'dropout': 0.0,
                    'residual': False,
                    'separable': True,
                    'se': True,
                    'se_context_size': -1,
                },
                {
                    'filters': model_defaults['enc_hidden'],
                    'repeat': 1,
                    'kernel': [1],
                    'stride': [1],
                    'dilation': [1],
                    'dropout': 0.0,
                    'residual': False,
                    'separable': True,
                    'se': True,
                    'se_context_size': -1,
                },
                {
                    'filters': model_defaults['enc_hidden'],
                    'repeat': 1,
                    'kernel': [1],
                    'stride': [1],
                    'dilation': [1],
                    'dropout': 0.0,
                    'residual': False,
                    'separable': True,
                    'se': True,
                    'se_context_size': -1,
                },
            ],
        },
    }

    spec_augment = {
        '_target_': 'nemo.collections.asr.modules.SpectrogramAugmentation',
        'freq_masks': 0,
        'time_masks': 0,
        'freq_width': 16,
        'time_width': 0.05,
    }

    masking = {
        "_target_": "nemo.collections.asr.modules.RandomBlockMasking",
        "block_size": 40,  # for pre_conv masking, 10ms per frame, 400ms per block with block_size=40
        "mask_prob": 0.01,  # for allow_overlap=True, this means the mask prob for each frame; otherwise it means the overall masked proportion
        "feat_in": preprocessor["features"],
        "freeze": True,
        "allow_overlap": True,
    }

    quantizer = {
        "_target_": "nemo.collections.asr.modules.RandomProjectionVectorQuantizer",
        "feat_in": preprocessor["features"],
        "code_dim": model_defaults["code_dim"],
        "num_books": model_defaults["num_books"],
        "num_classes": model_defaults["num_classes"],
        "dist_fn": "l2",  # choices=["l2", "cosine"]
        "freeze": True,
        "squeeze_single": model_defaults["squeeze_single"],
        "combine_time_steps": model_defaults["subsampling_factor"],  # conformer sub-sampling ratio
    }

    decoder = {
        "_target_": "nemo.collections.asr.modules.MultiSoftmaxDecoder",
        "feat_in": model_defaults["enc_hidden"],
        "num_classes": model_defaults["num_classes"],
        "num_decoders": model_defaults["num_books"],
        "squeeze_single": model_defaults["squeeze_single"],
        "use_bias": True,
    }

    loss = {
        "_target_": "nemo.collections.asr.losses.MultiMLMLoss",
        "combine_time_steps": model_defaults[
            "subsampling_factor"
        ],  # conformer sub-sampling ratio for 'pre_conv', 1 for 'post_conv'
        "mask_threshold": 0.8,
        "num_decoders": model_defaults["num_books"],
        "squeeze_single": model_defaults["squeeze_single"],
    }

    optim = {
        "name": "adamw",
        "lr": 5.0,
        # optimizer arguments
        "betas": [0.9, 0.98],
        "weight_decay": 1e-3,
    }

    model_config = DictConfig(
        {
            "preprocessor": DictConfig(preprocessor),
            "spec_augment": DictConfig(spec_augment),
            'model_defaults': DictConfig(model_defaults),
            "masking": DictConfig(masking),
            "quantizer": DictConfig(quantizer),
            "encoder": DictConfig(encoder),
            "decoder": DictConfig(decoder),
            "loss": DictConfig(loss),
            "optim": DictConfig(optim),
        }
    )
    ssl_model = EncDecDenoiseMaskedTokenPredModel(cfg=model_config)
    return ssl_model


class TestSSLModel:
    @pytest.mark.unit
    def test_constructor(self, ssl_model):
        confdict = ssl_model.to_config_dict()
        instance2 = SpeechEncDecSelfSupervisedModel.from_config_dict(confdict)
        assert isinstance(instance2, SpeechEncDecSelfSupervisedModel)

    @pytest.mark.unit
    def test_contr_nonquant(self, ssl_model):
        modelConfig_contr_nonquant = ssl_model.to_config_dict()

        loss_list_contr_nonquant = dict(modelConfig_contr_nonquant['loss_list'])
        del loss_list_contr_nonquant['mlm']

        loss_list_contr_nonquant['contr']['loss']['quantized_targets'] = False

        modelConfig_contr_nonquant['loss_list'] = DictConfig(loss_list_contr_nonquant)

        ssl_model = SpeechEncDecSelfSupervisedModel(cfg=modelConfig_contr_nonquant)

        input_signal = torch.randn(size=(4, 64000))
        length = torch.randint(low=48000, high=64000, size=[4])

        with torch.no_grad():
            spectrograms, spec_masks, encoded, encoded_len = ssl_model.forward(
                input_signal=input_signal, input_signal_length=length
            )

            loss_value, loss_val_dict = ssl_model.decoder_loss_step(spectrograms, spec_masks, encoded, encoded_len)

        assert len(loss_val_dict) == 1

    @pytest.mark.unit
    def test_contr_mlm(self, ssl_model):

        input_signal = torch.randn(size=(4, 64000))
        length = torch.randint(low=48000, high=64000, size=[4])

        with torch.no_grad():
            spectrograms, spec_masks, encoded, encoded_len = ssl_model.forward(
                input_signal=input_signal, input_signal_length=length
            )

        loss_value, loss_val_dict = ssl_model.decoder_loss_step(spectrograms, spec_masks, encoded, encoded_len)

        assert len(loss_val_dict) == 2

    @pytest.mark.unit
    def test_contr_mlm_multi(self, ssl_model):
        modelConfig_contr_mlm_multi = ssl_model.to_config_dict()

        model_defaults = modelConfig_contr_mlm_multi['model_defaults']

        loss_list_contr_mlm_multi = dict(modelConfig_contr_mlm_multi['loss_list'])
        loss_list_contr_mlm_multi['mlm_2'] = {
            'decoder': {
                '_target_': 'nemo.collections.asr.modules.ConvASRDecoder',
                'feat_in': model_defaults['enc_hidden'],
                'num_classes': 4096,
            },
            'loss': {'_target_': 'nemo.collections.asr.losses.MLMLoss', 'combine_time_steps': 1},
            'output_from_layer': "encoder.0",
            'targets_from_loss': "contr",
        }
        loss_list_contr_mlm_multi['mlm_3'] = {
            'decoder': {
                '_target_': 'nemo.collections.asr.modules.ConvASRDecoder',
                'feat_in': model_defaults['enc_hidden'],
                'num_classes': 4096,
            },
            'loss': {'_target_': 'nemo.collections.asr.losses.MLMLoss', 'combine_time_steps': 1},
            'output_from_layer': "encoder.1",
            'targets_from_loss': "contr",
        }
        modelConfig_contr_mlm_multi['loss_list'] = DictConfig(loss_list_contr_mlm_multi)

        ssl_model = SpeechEncDecSelfSupervisedModel(cfg=modelConfig_contr_mlm_multi)

        input_signal = torch.randn(size=(4, 64000))
        length = torch.randint(low=48000, high=64000, size=[4])

        with torch.no_grad():
            spectrograms, spec_masks, encoded, encoded_len = ssl_model.forward(
                input_signal=input_signal, input_signal_length=length
            )

            loss_value, loss_val_dict = ssl_model.decoder_loss_step(spectrograms, spec_masks, encoded, encoded_len)

        assert len(loss_val_dict) == 4


class TestDenoiseMLMSSLModel:
    @pytest.mark.unit
    def test_forward(self, denoise_mlm_ssl_model):
        input_signal = torch.randn(size=(4, 64000))
        input_length = torch.randint(low=48000, high=64000, size=[4])
        noise = 0.1 * torch.ones_like(input_signal)
        noisy_input_signal = input_signal + noise
        noisy_input_length = input_length
        with torch.no_grad():
            with typecheck.disable_checks():
                log_probs, encoded_len, masks, tokens = denoise_mlm_ssl_model.forward(
                    input_signal=input_signal,
                    input_signal_length=input_length,
                    noisy_input_signal=noisy_input_signal,
                    noisy_input_signal_length=noisy_input_length,
                )

        assert log_probs.size(0) == 4
        assert log_probs.size(2) == denoise_mlm_ssl_model.cfg.model_defaults.num_classes
        assert encoded_len.size(0) == 4
        assert masks.size(0) == 4
        assert tokens.size(0) == 4
        assert masks.sum() == 0.0  # no mask should be applied to the input by default

    @pytest.mark.unit
    def test_forward_masked(self, denoise_mlm_ssl_model: EncDecDenoiseMaskedTokenPredModel):
        input_signal = torch.randn(size=(4, 64000))
        input_length = torch.randint(low=48000, high=64000, size=[4])
        noise = 0.1 * torch.ones_like(input_signal)
        noisy_input_signal = input_signal + noise
        noisy_input_length = input_length

        with torch.no_grad():
            with typecheck.disable_checks():
                log_probs, encoded_len, masks, tokens = denoise_mlm_ssl_model.forward(
                    input_signal=input_signal,
                    input_signal_length=input_length,
                    noisy_input_signal=noisy_input_signal,
                    noisy_input_signal_length=noisy_input_length,
                    apply_mask=True,
                )

        loss_value = denoise_mlm_ssl_model.loss(
            masks=masks, decoder_outputs=log_probs, targets=tokens, decoder_lengths=encoded_len
        )

        assert log_probs.size(0) == 4
        assert log_probs.size(2) == denoise_mlm_ssl_model.cfg.model_defaults.num_classes
        assert encoded_len.size(0) == 4
        assert masks.size(0) == 4
        assert tokens.size(0) == 4
        assert masks.sum() > 0.0  # mask should be applied to the input
        assert not torch.isnan(loss_value)



# Copyright (c) 2022, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# https://github.com/NVIDIA/Megatron-LM/blob/main/megatron/data/autoaugment.py

import random

import numpy as np
from PIL import Image, ImageEnhance, ImageOps

_MAX_LEVEL = 10  # Maximum integer strength of an augmentation, if applicable.


class ImageNetPolicy:
    """Definition of an ImageNetPolicy.

    Implements a fixed AutoAugment data augmentation policy targeted at
    ImageNet training by randomly applying at runtime one of the 25 pre-defined
    data augmentation sub-policies provided in Reference [1].

    Usage example as a Pytorch Transform:
    >>> transform=transforms.Compose([transforms.Resize(256),
    >>>                               ImageNetPolicy(),
    >>>                               transforms.ToTensor()])
    """

    def __init__(self, fillcolor=(128, 128, 128)):
        """Initialize an ImageNetPolicy.

        Args:
            fillcolor (tuple): RGB color components of the color to be used for
            filling when needed (default: (128, 128, 128), which
            corresponds to gray).
        """
        # Instantiate a list of sub-policies.
        # Each entry of the list is a SubPolicy which consists of
        # two augmentation operations,
        # each of those parametrized as operation, probability, magnitude.
        # Those two operations are applied sequentially on the image upon call.
        self.policies = [
            SubPolicy("posterize", 0.4, 8, "rotate", 0.6, 9, fillcolor),
            SubPolicy("solarize", 0.6, 5, "autocontrast", 0.6, 5, fillcolor),
            SubPolicy("equalize", 0.8, 8, "equalize", 0.6, 3, fillcolor),
            SubPolicy("posterize", 0.6, 7, "posterize", 0.6, 6, fillcolor),
            SubPolicy("equalize", 0.4, 7, "solarize", 0.2, 4, fillcolor),
            SubPolicy("equalize", 0.4, 4, "rotate", 0.8, 8, fillcolor),
            SubPolicy("solarize", 0.6, 3, "equalize", 0.6, 7, fillcolor),
            SubPolicy("posterize", 0.8, 5, "equalize", 1.0, 2, fillcolor),
            SubPolicy("rotate", 0.2, 3, "solarize", 0.6, 8, fillcolor),
            SubPolicy("equalize", 0.6, 8, "posterize", 0.4, 6, fillcolor),
            SubPolicy("rotate", 0.8, 8, "color", 0.4, 0, fillcolor),
            SubPolicy("rotate", 0.4, 9, "equalize", 0.6, 2, fillcolor),
            SubPolicy("equalize", 0.0, 7, "equalize", 0.8, 8, fillcolor),
            SubPolicy("invert", 0.6, 4, "equalize", 1.0, 8, fillcolor),
            SubPolicy("color", 0.6, 4, "contrast", 1.0, 8, fillcolor),
            SubPolicy("rotate", 0.8, 8, "color", 1.0, 2, fillcolor),
            SubPolicy("color", 0.8, 8, "solarize", 0.8, 7, fillcolor),
            SubPolicy("sharpness", 0.4, 7, "invert", 0.6, 8, fillcolor),
            SubPolicy("shearX", 0.6, 5, "equalize", 1.0, 9, fillcolor),
            SubPolicy("color", 0.4, 0, "equalize", 0.6, 3, fillcolor),
            SubPolicy("equalize", 0.4, 7, "solarize", 0.2, 4, fillcolor),
            SubPolicy("solarize", 0.6, 5, "autocontrast", 0.6, 5, fillcolor),
            SubPolicy("invert", 0.6, 4, "equalize", 1.0, 8, fillcolor),
            SubPolicy("color", 0.6, 4, "contrast", 1.0, 8, fillcolor),
            SubPolicy("equalize", 0.8, 8, "equalize", 0.6, 3, fillcolor),
        ]

    def __call__(self, img):
        """Define call method for ImageNetPolicy class."""
        policy_idx = random.randint(0, len(self.policies) - 1)
        return self.policies[policy_idx](img)

    def __repr__(self):
        """Define repr method for ImageNetPolicy class."""
        return "ImageNetPolicy"


class SubPolicy:
    """Definition of a SubPolicy.

    A SubPolicy consists of two augmentation operations,
    each of those parametrized as operation, probability, magnitude.
    The two operations are applied sequentially on the image upon call.
    """

    def __init__(
        self, operation1, probability1, magnitude_idx1, operation2, probability2, magnitude_idx2, fillcolor,
    ):
        """Initialize a SubPolicy.

        Args:
            operation1 (str): Key specifying the first augmentation operation.
            There are fourteen key values altogether (see supported_ops below
            listing supported operations). probability1 (float): Probability
            within [0., 1.] of applying the first augmentation operation.
            magnitude_idx1 (int): Integer specifiying the strength of the first
            operation as an index further used to derive the magnitude from a
            range of possible values.
            operation2 (str): Key specifying the second augmentation operation.
            probability2 (float): Probability within [0., 1.] of applying the
            second augmentation operation.
            magnitude_idx2 (int): Integer specifiying the strength of the
            second operation as an index further used to derive the magnitude
            from a range of possible values.
            fillcolor (tuple): RGB color components of the color to be used for
            filling.
        Returns:
        """
        # List of supported operations for operation1 and operation2.
        supported_ops = [
            "shearX",
            "shearY",
            "translateX",
            "translateY",
            "rotate",
            "color",
            "posterize",
            "solarize",
            "contrast",
            "sharpness",
            "brightness",
            "autocontrast",
            "equalize",
            "invert",
        ]
        assert (operation1 in supported_ops) and (
            operation2 in supported_ops
        ), "SubPolicy:one of oper1 or oper2 refers to an unsupported operation."

        assert (
            0.0 <= probability1 <= 1.0 and 0.0 <= probability2 <= 1.0
        ), "SubPolicy: prob1 and prob2 should be within [0., 1.]."

        assert (
            isinstance(magnitude_idx1, int) and 0 <= magnitude_idx1 <= 10
        ), "SubPolicy: idx1 should be specified as an integer within [0, 10]."

        assert (
            isinstance(magnitude_idx2, int) and 0 <= magnitude_idx2 <= 10
        ), "SubPolicy: idx2 should be specified as an integer within [0, 10]."

        # Define a dictionary where each key refers to a specific type of
        # augmentation and the corresponding value is a range of ten possible
        # magnitude values for that augmentation.
        num_levels = _MAX_LEVEL + 1
        ranges = {
            "shearX": np.linspace(0, 0.3, num_levels),
            "shearY": np.linspace(0, 0.3, num_levels),
            "translateX": np.linspace(0, 150 / 331, num_levels),
            "translateY": np.linspace(0, 150 / 331, num_levels),
            "rotate": np.linspace(0, 30, num_levels),
            "color": np.linspace(0.0, 0.9, num_levels),
            "posterize": np.round(np.linspace(8, 4, num_levels), 0).astype(np.int64),
            "solarize": np.linspace(256, 0, num_levels),  # range [0, 256]
            "contrast": np.linspace(0.0, 0.9, num_levels),
            "sharpness": np.linspace(0.0, 0.9, num_levels),
            "brightness": np.linspace(0.0, 0.9, num_levels),
            "autocontrast": [0] * num_levels,  # This augmentation doesn't use magnitude parameter.
            "equalize": [0] * num_levels,  # This augmentation doesn't use magnitude parameter.
            "invert": [0] * num_levels,  # This augmentation doesn't use magnitude parameter.
        }

        def rotate_with_fill(img, magnitude):
            """Define rotation transformation with fill.

            The input image is first rotated, then it is blended together with
            a gray mask of the same size. Note that fillcolor as defined
            elsewhere in this module doesn't apply here.

            Args:
                magnitude (float): rotation angle in degrees.
            Returns:
                rotated_filled (PIL Image): rotated image with gray filling for
                disoccluded areas unveiled by the rotation.
            """
            rotated = img.convert("RGBA").rotate(magnitude)
            rotated_filled = Image.composite(rotated, Image.new("RGBA", rotated.size, (128,) * 4), rotated)
            return rotated_filled.convert(img.mode)

        # Define a dictionary of augmentation functions where each key refers
        # to a specific type of augmentation and the corresponding value defines
        # the augmentation itself using a lambda function.
        # pylint: disable=unnecessary-lambda
        func_dict = {
            "shearX": lambda img, magnitude: img.transform(
                img.size,
                Image.AFFINE,
                (1, magnitude * random.choice([-1, 1]), 0, 0, 1, 0),
                Image.BICUBIC,
                fillcolor=fillcolor,
            ),
            "shearY": lambda img, magnitude: img.transform(
                img.size,
                Image.AFFINE,
                (1, 0, 0, magnitude * random.choice([-1, 1]), 1, 0),
                Image.BICUBIC,
                fillcolor=fillcolor,
            ),
            "translateX": lambda img, magnitude: img.transform(
                img.size,
                Image.AFFINE,
                (1, 0, magnitude * img.size[0] * random.choice([-1, 1]), 0, 1, 0,),
                fillcolor=fillcolor,
            ),
            "translateY": lambda img, magnitude: img.transform(
                img.size,
                Image.AFFINE,
                (1, 0, 0, 0, 1, magnitude * img.size[1] * random.choice([-1, 1]),),
                fillcolor=fillcolor,
            ),
            "rotate": lambda img, magnitude: rotate_with_fill(img, magnitude),
            "color": lambda img, magnitude: ImageEnhance.Color(img).enhance(1 + magnitude * random.choice([-1, 1])),
            "posterize": lambda img, magnitude: ImageOps.posterize(img, magnitude),
            "solarize": lambda img, magnitude: ImageOps.solarize(img, magnitude),
            "contrast": lambda img, magnitude: ImageEnhance.Contrast(img).enhance(
                1 + magnitude * random.choice([-1, 1])
            ),
            "sharpness": lambda img, magnitude: ImageEnhance.Sharpness(img).enhance(
                1 + magnitude * random.choice([-1, 1])
            ),
            "brightness": lambda img, magnitude: ImageEnhance.Brightness(img).enhance(
                1 + magnitude * random.choice([-1, 1])
            ),
            "autocontrast": lambda img, magnitude: ImageOps.autocontrast(img),
            "equalize": lambda img, magnitude: ImageOps.equalize(img),
            "invert": lambda img, magnitude: ImageOps.invert(img),
        }

        # Store probability, function and magnitude of the first augmentation
        # for the sub-policy.
        self.probability1 = probability1
        self.operation1 = func_dict[operation1]
        self.magnitude1 = ranges[operation1][magnitude_idx1]

        # Store probability, function and magnitude of the second augmentation
        # for the sub-policy.
        self.probability2 = probability2
        self.operation2 = func_dict[operation2]
        self.magnitude2 = ranges[operation2][magnitude_idx2]

    def __call__(self, img):
        """Define call method for SubPolicy class."""
        # Randomly apply operation 1.
        if random.random() < self.probability1:
            img = self.operation1(img, self.magnitude1)

        # Randomly apply operation 2.
        if random.random() < self.probability2:
            img = self.operation2(img, self.magnitude2)

        return img
